import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import datetime
from datetime import date, timedelta
import calendar
from Text2JSON.entity_named_recog.entity_utils import *
from Text2JSON.entity_named_recog.year_entity_recognition import year_absolute_comfirm, year_relative_comfirm


def quart_recon(line, placeholders_list: dict):
    return quarter_relative_absolute_recognition(line, placeholders_list)


def quarter_relative_absolute_recognition(line, placeholders_list: dict):
    cur_quarter = ['最近一个季度', '近一个季度', '最近1个季度', '近1个季度', '近季度', '本季度', '这个季度']
    next_quarter = ['下1个季度', '下一个季度', '下季度', '下个季度', '下季度', '下一季度', '下1季度', '一个季度后', '1个季度后']
    last_quarter = ['上1个季度', '上一个季度', '上季度', '上个季度', '上季度', '上一季度', '上1季度', '一个季度前', '1个季度前']
    first_quarter = ['第一季度', '第1季度', '一季度', '1季度']  # 最长匹配，最长的放在前面
    sec_quarter = ['第二季度', '第2季度', '二季度', '2季度']
    third_quarter = ['第三季度', '第3季度', '三季度', '3季度']
    fourth_quarter = ['第四季度', '第4季度', '四季度', '4季度', '最后一个季度', '最后1个季度']
    first_half_year = ['上半年']
    sec_half_year = ['下半年']
    flag = True
    while flag:
        flag = False
        for quarter in cur_quarter:
            if quarter in line:
                index = line.index(quarter)
                first_line = line[0: index]
                sec_line = line[index + len(quarter):]

                now = date.today()
                # 季度开始月份, 季度结束月份, 季度开始时间, 季度结束时间
                month_start = (now.month - 1) - (now.month - 1) % 3 + 1
                month_end = month_start + 2
                cur_quarter_start = datetime.datetime(now.year, month_start, 1)
                cur_quarter_end = datetime.datetime(now.year, month_end, calendar.monthrange(now.year, month_end)[1])
                time_str = '“' + date_to_str(cur_quarter_start) + '”' + '到' + '“' + date_to_str(cur_quarter_end) + '”'
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [cur_quarter_start, cur_quarter_end])
                line = first_line + placeholder + sec_line
                flag = True
                break

        for quarter in next_quarter:
            if quarter in line:
                index = line.index(quarter)
                first_line = line[0: index]
                sec_line = line[index + len(quarter):]

                now = date.today()
                # 思路，先获取本季度的结束月份最后天，日+1，再获取下一个季度的开始和结束时间
                month_start = (now.month - 1) - (now.month - 1) % 3 + 1
                month_end = month_start + 2
                this_quarter_end = datetime.datetime(now.year, month_end, calendar.monthrange(now.year, month_end)[1])

                # 季度开始月份, 季度结束月份, 季度开始时间, 季度结束时间
                next_quarter_temp = (this_quarter_end + timedelta(days=1))
                next_quarter_start_year = next_quarter_temp.year
                next_quarter_start_month = (next_quarter_temp.month - 1) - (next_quarter_temp.month - 1) % 3 + 1
                next_quarter_end_month = next_quarter_start_month + 2

                next_quarter_start = datetime.datetime(next_quarter_start_year, next_quarter_start_month, 1)
                next_quarter_end = datetime.datetime(next_quarter_start_year, next_quarter_end_month,
                                                     calendar.monthrange(next_quarter_start_year,
                                                                         next_quarter_end_month)[1])

                time_str = '“' + date_to_str(next_quarter_start) + '”' + '到' + '“' + date_to_str(next_quarter_end) + '”'
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [next_quarter_start, next_quarter_end])
                line = first_line + placeholder + sec_line
                flag = True
                break

        for quarter in last_quarter:
            if quarter in line:
                index = line.index(quarter)
                first_line = line[0: index]
                sec_line = line[index + len(quarter):]

                now = date.today()
                # 思路，先获取本季度的开始月份第一天，日-1，再获取上一个季度的开始和结束时间
                month_start = (now.month - 1) - (now.month - 1) % 3 + 1
                this_quarter_start = datetime.datetime(now.year, month_start, 1)

                # 季度开始月份, 季度结束月份, 季度开始时间, 季度结束时间
                last_quarter_temp = (this_quarter_start + timedelta(days=-1))
                last_quarter_start_year = last_quarter_temp.year
                last_quarter_start_month = (last_quarter_temp.month - 1) - (last_quarter_temp.month - 1) % 3 + 1
                pre_quarter_end_month = last_quarter_start_month + 2

                last_quarter_start = datetime.datetime(last_quarter_start_year, last_quarter_start_month, 1)
                last_quarter_end = datetime.datetime(last_quarter_start_year, pre_quarter_end_month,
                                                     calendar.monthrange(last_quarter_start_year,
                                                                         pre_quarter_end_month)[1])

                time_str = '“' + date_to_str(last_quarter_start) + '”' + '到' + '“' + date_to_str(last_quarter_end) + '”'
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [last_quarter_start, last_quarter_end])
                line = first_line + placeholder + sec_line
                flag = True
                break

        for quarter in first_quarter:
            if quarter in line:
                index = line.index(quarter)
                first_line = line[0: index]
                sec_line = line[index + len(quarter):]

                now = date.today()
                # 确定年份
                first_line, year = year_absolute_comfirm(first_line, placeholders_list)
                if year is None:
                    first_line, year = year_relative_comfirm(first_line, placeholders_list)

                year = now.year if year is None else year

                this_quarter_start = datetime.datetime(year, 1, 1)
                this_quarter_end = datetime.datetime(year, 3, 31)
                time_str = '“' + date_to_str(this_quarter_start) + '”' + '到' + '“' + date_to_str(this_quarter_end) + '”'
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [this_quarter_start, this_quarter_end])
                line = first_line + placeholder + sec_line
                flag = True
                break

        for quarter in sec_quarter:
            if quarter in line:
                index = line.index(quarter)
                first_line = line[0: index]
                sec_line = line[index + len(quarter):]

                now = date.today()
                # 确定年份
                first_line, year = year_absolute_comfirm(first_line, placeholders_list)
                if year is None:
                    first_line, year = year_relative_comfirm(first_line, placeholders_list)
                year = now.year if year is None else year
                this_quarter_start = datetime.datetime(year, 4, 1)
                this_quarter_end = datetime.datetime(year, 6, 30)
                time_str = '“' + date_to_str(this_quarter_start) + '”' + '到' + '“' + date_to_str(this_quarter_end) + '”'
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [this_quarter_start, this_quarter_end])
                line = first_line + placeholder + sec_line
                flag = True
                break

        for quarter in third_quarter:
            if quarter in line:
                index = line.index(quarter)
                first_line = line[0: index]
                sec_line = line[index + len(quarter):]

                now = date.today()
                # 确定年份
                first_line, year = year_absolute_comfirm(first_line, placeholders_list)
                if year is None:
                    first_line, year = year_relative_comfirm(first_line, placeholders_list)
                year = now.year if year is None else year
                this_quarter_start = datetime.datetime(year, 7, 1)
                this_quarter_end = datetime.datetime(year, 9, 30)
                time_str = '“' + date_to_str(this_quarter_start) + '”' + '到' + '“' + date_to_str(this_quarter_end) + '”'
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [this_quarter_start, this_quarter_end])
                line = first_line + placeholder + sec_line
                flag = True
                break

        for quarter in fourth_quarter:
            if quarter in line:
                index = line.index(quarter)
                first_line = line[0: index]
                sec_line = line[index + len(quarter):]

                now = date.today()
                # 确定年份
                first_line, year = year_absolute_comfirm(first_line, placeholders_list)
                if year is None:
                    first_line, year = year_relative_comfirm(first_line, placeholders_list)
                year = now.year if year is None else year
                this_quarter_start = datetime.datetime(year, 10, 1)
                this_quarter_end = datetime.datetime(year, 12, 31)
                time_str = '“' + date_to_str(this_quarter_start) + '”' + '到' + '“' + date_to_str(this_quarter_end) + '”'
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [this_quarter_start, this_quarter_end])
                line = first_line + placeholder + sec_line
                flag = True
                break

        for first_half in first_half_year:
            if first_half in line:
                index = line.index(first_half)
                first_line = line[0: index]
                sec_line = line[index + len(first_half):]

                now = date.today()
                # 确定年份
                first_line, year = year_absolute_comfirm(first_line, placeholders_list)
                if year is None:
                    first_line, year = year_relative_comfirm(first_line, placeholders_list)
                year = now.year if year is None else year
                this_half_start = datetime.datetime(year, 1, 1)
                this_half_end = datetime.datetime(year, 6, 30)
                time_str = '“' + date_to_str(this_half_start) + '”' + '到' + '“' + date_to_str(this_half_end) + '”'
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [this_half_start, this_half_end])
                line = first_line + placeholder + sec_line
                flag = True
                break

        for sec_half in sec_half_year:
            if sec_half in line:
                index = line.index(sec_half)
                first_line = line[0: index]
                sec_line = line[index + len(sec_half):]

                now = date.today()
                # 确定年份
                first_line, year = year_absolute_comfirm(first_line, placeholders_list)
                if year is None:
                    first_line, year = year_relative_comfirm(first_line, placeholders_list)
                year = now.year if year is None else year
                this_half_start = datetime.datetime(year, 7, 1)
                this_half_end = datetime.datetime(year, 12, 31)
                time_str = '“' + date_to_str(this_half_start) + '”' + '到' + '“' + date_to_str(this_half_end) + '”'
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [this_half_start, this_half_end])
                line = first_line + placeholder + sec_line
                flag = True
                break

    return line


if __name__ == '__main__':
    ''
    # example1 = '12年第1季度，16年第二季度，2017年第三季度，2019年第四季度，最近一个季度，上季度，下季度'
    # example2 = '今年第一季度，去年第二季度，明年第三季度，第四季度'
    # example3 = '17年上半年，下半年，今年下半年，明年上半年'
    # sentence, placeholders = quarter_relatice_absolute_recon(example1)
    # for token, holder in placeholders:
    #     sentence = sentence.replace(holder, token)
    # print(sentence)
    # sentence, placeholders = quarter_relatice_absolute_recon(example2)
    # for token, holder in placeholders:
    #     sentence = sentence.replace(holder, token)
    # print(sentence)
    # sentence, placeholders = quarter_relatice_absolute_recon(example3)
    # for token, holder in placeholders:
    #     sentence = sentence.replace(holder, token)
    # print(sentence)



import random


def date_to_str(tem_date):
    year = str(tem_date.year)
    month = str(tem_date.month) if len(str(tem_date.month)) == 2 else '0' + str(tem_date.month)
    day = str(tem_date.day) if len(str(tem_date.day)) == 2 else '0' + str(tem_date.day)
    return year + '年' + month + '月' + day + '日'


def time_to_str(tem_date):
    year = str(tem_date.year)
    month = str(tem_date.month) if len(str(tem_date.month)) == 2 else '0' + str(tem_date.month)
    day = str(tem_date.day) if len(str(tem_date.day)) == 2 else '0' + str(tem_date.day)
    hour = str(tem_date.hour) if len(str(tem_date.hour)) == 2 else '0' + str(tem_date.hour)
    minute = str(tem_date.minute) if len(str(tem_date.minute)) == 2 else '0' + str(tem_date.minute)
    return year + '年' + month + '月' + day + '日' + hour + ':' + minute


def random_str(slen=15):
    seed = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!@#$%^&*()_+="
    sa = []
    for i in range(slen):
        sa.append(random.choice(seed))
    return ''.join(sa)



import sys
from os.path import dirname, abspath

path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import json
import numpy as np
import os
from Text2JSON.model.base_model import BaseModel
from Text2JSON.featurizer.schema_featurizer import SchemaDataset
from Text2JSON.featurizer.column_featurizer import ColumnDataset
from Text2JSON.featurizer.Input_SQL import SQLExample
from Text2JSON.train.utils import dump_model, load_model

op_map = {'': 0, 'eq': 1, 'not': 2, 'gt': 3, 'lt': 4, 'gte': 5, 'lte': 6, 'like': 7,
          'isnull': 8, 'notnull': 9, 'btw': 10, 'in': 11}
agg_map = {'': 0, 'eq': 1, 'not': 2, 'gt': 3, 'lt': 4, 'gte': 5, 'lte': 6, 'like': 7,
           'isnull': 8, 'notnull': 9, 'btw': 10, 'in': 11}

agg_map_reverse = {0: '', 1: 'eq', 2: 'not', 3: 'gt', 4: 'lt', 5: 'gte', 6: 'lte', 7: 'like', 8: 'isnull', 9: 'notnull',
                   10: 'btw', 11: 'in'}
op_map_reverse = {0: '', 1: 'eq', 2: 'not', 3: 'gt', 4: 'lt', 5: 'gte', 6: 'lte', 7: 'like', 8: 'isnull', 9: 'notnull',
                  10: 'btw', 11: 'in'}


class request_template(object):
    def __init__(self, method, url, params):
        self.method = method
        self.url = url
        self.params = params

    def to_dict(self):
        return {
            'method': self.method,
            'url': self.url,
            'params': self.params
        }


class GenJSON(object):
    def __init__(self, config, schema_data, column_data, schema, model1, model2):
        self.config = config
        self.schema_data = schema_data
        self.column_data = column_data
        self.schema = schema
        self.model1 = model1
        self.model2 = model2

    def gen_json(self, question, first_json=1, sec_json=1, qid=0, size=50):
        self.config['schema_evaluator_batch_size'] = size
        self.config['column_evaluator_batch_size'] = size
        example = self.question_pre_handle(question, qid)
        self.schema_data.load_for_predict(example, True)
        schema_model_outputs = self.model1.dataset_inference(self.schema_data)

        # SQL1 relative
        first_table_relevant_prob = np.exp(schema_model_outputs[0]['schema_first_sim'][:, 0])
        first_table_id_prob_exp = sorted(enumerate(first_table_relevant_prob), key=lambda x: x[1], reverse=True)
        first_table_rel = []
        for table_index, table_pro in first_table_id_prob_exp:
            if table_pro >= float(self.config['schema_threshold']):
                first_table_rel.append(table_index)
                if len(first_table_rel) == int(self.config['schema_remain']):
                    break

        # SQL2 relative
        sec_table_relevant_prob = np.exp(schema_model_outputs[0]['schema_sec_sim'][:, 0])
        sec_table_id_prob_exp = sorted(enumerate(sec_table_relevant_prob), key=lambda x: x[1], reverse=True)
        sec_table_rel = []
        for table_index, table_pro in sec_table_id_prob_exp:
            if table_pro >= float(self.config['schema_threshold']):
                sec_table_rel.append(table_index)
                if len(first_table_rel) == int(self.config['schema_remain']):
                    break

        result = None
        # 先不考虑剪枝
        if len(first_table_rel) == 0:
            result = {'qid': qid, 'question': question, 'query': []}
        # 两句SQL
        else:
            # zip SQL1 and SQL2
            example.first_schema_id = [self.schema.table_global_to_local[index] for index in first_table_rel]
            example.sec_schema_id = [self.schema.table_global_to_local[index] for index in sec_table_rel]
            self.column_data.load_for_predict(example, True)

            # 用于标识两个语句的关系
            # model output SQL
            column_model_outputs = self.model2.dataset_inference(self.column_data)
            for sql1_index, first_schema_id in enumerate(first_table_rel):
                first_relation, sec_relation = None, None
                if sql1_index == int(first_json):
                    break
                first_schema_id = self.schema.table_global_to_local[first_schema_id]
                input_feature = self.column_data.input_features[sql1_index]
                model_output = column_model_outputs[sql1_index]
                first_request, first_relation = self.sql_parser(self.model2, first_schema_id, input_feature,
                                                                model_output, 0)
                if len(sec_table_rel) == 0:
                    result = {'qid': qid, 'question': question, 'query': [first_request.to_dict()]}
                    continue
                for sql2_index, sec_schema_id in enumerate(sec_table_rel):
                    if sql2_index == int(sec_json):
                        break
                    sec_schema_id = self.schema.table_global_to_local[sec_schema_id]
                    input_feature = self.column_data.input_features[len(first_table_rel) + sql2_index]
                    model_output = column_model_outputs[len(first_table_rel) + sql2_index]
                    sec_request, sec_relation = self.sql_parser(self.model2, sec_schema_id, input_feature, model_output,
                                                                1)

                    # 剪枝
                    if first_request.url == sec_request.url:
                        # 只要第一句
                        result = {'qid': qid, 'question': question, 'query': [first_request.to_dict()]}
                    elif first_relation == 0 and sec_relation == 0:
                        # 只要第一句
                        result = {'qid': qid, 'question': question, 'query': [first_request.to_dict()]}
                    else:
                        # 两句都要
                        result = {'qid': qid, 'question': question,
                                  'query': [first_request.to_dict(), sec_request.to_dict()]}
        return result

    def question_pre_handle(self, line, qid):
        """
            qid
            question
            table_id
            tokens
            word_to_char_start
        """

        example = SQLExample()
        example.qid = qid
        example.table_id = None
        line = line.replace(' ', '')
        tokens = [char for char in line]
        example.tokens = tokens
        example.question = line

        space_list = [i for i, x in enumerate(' '.join(tokens)) if x == ' ']
        word_to_char_start = [0]
        word_to_char_start.extend([space_list[index] + 1 for index in range(0, len(space_list) - 1)])
        example.word_to_char_start = word_to_char_start
        return example

    def generate_json(self, model: BaseModel, input_feature, model_output, sql_seq=0):
        """
            predict sql on dataset
            dataset: 需要预测的数据
            model_outputs: 数据经过模型预测后的输出
        """
        stop_word_list = ['【', '】', '[', ']', '"', "'", '‘', '’', '“', '，', '.', '。', ' ', '”']

        sel, agg, where, _, conditions, relationship = model. \
            parse_output_with_threshold(input_feature, model_output, sql_seq, float(self.config['column_threshold']))
        conditions_with_value_texts = []
        for wc in where:
            _, op, span_list = conditions[wc]
            # 通过subword_to_word获取对应的token的start_index, end_index
            value_span_text = []
            for se in span_list:
                # 获取是第几个词
                start_index = se[0]
                end_index = se[1]
                if start_index >= len(input_feature.subword_to_word[wc]):
                    start_index = len(input_feature.subword_to_word[wc]) - 1
                if end_index >= len(input_feature.subword_to_word[wc]):
                    end_index = len(input_feature.subword_to_word[wc]) - 1
                word_start = input_feature.subword_to_word[wc][start_index]
                word_end = input_feature.subword_to_word[wc][end_index] + 1
                span = ''.join(input_feature.tokens_common[word_start:word_end]).rstrip()
                for stop_word in stop_word_list:
                    if stop_word in span:
                        span.replace(stop_word, '')
                value_span_text.append(span)
            conditions_with_value_texts.append((wc, op, value_span_text))

        # 将agg,select, conditions封装到sql中
        return sel, agg, conditions_with_value_texts, relationship

    def parse_params(self, schema_id, sel, agg, conditions_with_value_texts):
        # 遍历SQL语句
        params = []
        select_params = []
        where_params = []

        # 这里可以进行剪枝，看看select 有没有出现在 where 中，或者反过来
        for index, sel_col_index in enumerate(sel):
            col_name_value = self.schema.extend_schema[schema_id]['header'][sel_col_index]
            if '&' not in col_name_value:
                '去除占位符'
                continue
            col_name = col_name_value.split('&')[0]
            col_value = col_name_value.split('&')[1]

            agg_name = agg_map_reverse[agg[index][1]]
            if agg_name == 'in' or agg_name == 'btw':
                col_value = '"' + col_value + '"'
            select_params.append((col_name, agg_name, col_value))

        for where_col_index, op_index, value_list in conditions_with_value_texts:
            col_name = self.schema.extend_schema[schema_id]['header'][where_col_index]
            op_name = op_map_reverse[op_index]
            if len(value_list) >= 2:
                for value_index, value in enumerate(value_list):
                    value = value_list[value_index]
                    value = value.replace('”', '')
                    value = value.replace('"', '')
                    value_list[value_index] = '"' + str(value) + '"'
                value = value_list
            else:
                value = value_list[0]
            where_params.append((col_name, op_name, value))

        # 合并同类项
        select_name_agg = {}
        select_name_value = {}
        for col_name, agg_name, col_value in select_params:
            if col_name not in select_name_agg.keys():
                select_name_agg[col_name] = [agg_name]
            else:
                select_name_agg[col_name].append(agg_name)
            if col_name not in select_name_value.keys():
                select_name_value[col_name] = [col_value]
            else:
                select_name_value[col_name].append(col_value)

        # 对agg进行剪枝, 选择可能性最大的agg
        for key in select_name_agg.keys():
            if len(select_name_agg[key]) == 1:
                continue
            else:
                agg_name_num = {}
                for agg_name in select_name_agg[key]:
                    if agg_name not in agg_name_num.keys():
                        agg_name_num[agg_name] = 1
                    else:
                        agg_name_num[agg_name] += 1
                if len(agg_name_num.keys()) == 1:
                    continue
                find_max = -1
                max_agg_name = None
                for agg_name in agg_name_num.keys():
                    if agg_name_num[agg_name] > find_max:
                        find_max = agg_name_num[agg_name]
                        max_agg_name = agg_name
                # 这里引入的特征
                if max_agg_name == 'eq' and len(select_name_value[key]) > 1:
                    select_name_agg[key] = ['in']
                else:
                    select_name_agg[key] = [max_agg_name]

        params = []
        # 解析成JSON格式
        for col_name in select_name_agg.keys():
            params.append({'name': col_name, 'option': select_name_agg[col_name][0],
                           'value': str(select_name_value[col_name]) if len(select_name_value[col_name]) > 1
                                                                        or select_name_agg[col_name][0] == 'in'
                           else str(select_name_value[col_name][0])})
        for col_name, op_name, value in where_params:
            params.append({'name': col_name, 'option': op_name, 'value': str([value]) if op_name == 'in' else [value]})

        return params

    def get_url_method(self, table_id):
        url = self.schema.extend_schema[table_id]['url']
        method = self.schema.url_method[url]
        return url, method

    def sql_parser(self, model: BaseModel, schema_id, input_feature, model_output, sql_seq):
        sel, agg, conditions_with_value_texts, relationship = self.generate_json(model, input_feature, model_output,
                                                                                 sql_seq)
        # parsing params
        params = self.parse_params(schema_id, sel, agg, conditions_with_value_texts)
        # get url and method
        url, method = self.get_url_method(schema_id)
        # request
        return request_template(method, url, params), self.first_or_sec(relationship)

    def first_or_sec(self, relationship):
        if len(relationship) == 0:
            return 0
        first = 0
        sec = 0
        for index, rel in relationship:
            if rel == 0:
                first += 1
            else:
                sec += 1
        return 0 if first > sec else 1


def load_test_data(data_path):
    question_list = []
    for line in open(data_path, encoding="utf8"):
        data = json.loads(line)
        question_list.append((data['qid'], data['question'].replace(' ', '')))
    return question_list


class Predict(object):
    def __init__(self):
        # 设置参数
        self.config = None
        # 词编码器
        self.tokenizer = None
        # schema特征提取器
        self.schema_featurizer = None
        # column特征提取器
        self.column_featurizer = None
        # 加载数据表
        self.schema = None
        # schema模型
        self.model1 = None
        # column模型
        self.model2 = None
        # 数据加载器
        self.schema_data = None
        self.column_data = None
        # 预测器
        self.prediction = None

    def set_params(self, config, tokenizer, schema_featurizer, column_featurizer, schema,
                   schema_model, column_model):
        # 设置参数
        self.config = config
        # 词编码器
        self.tokenizer = tokenizer
        # schema特征提取器
        self.schema_featurizer = schema_featurizer
        # column特征提取器
        self.column_featurizer = column_featurizer
        # 加载数据表
        self.schema = schema
        # 选择模型, 加载模型
        self.model1 = schema_model
        self.model2 = column_model

        # 数据加载器
        self.schema_data = SchemaDataset(self.config, self.schema_featurizer, self.schema)
        self.column_data = ColumnDataset(self.config, self.column_featurizer, self.schema)

        # 预测器
        self.prediction = GenJSON(self.config, self.schema_data, self.column_data, self.schema, self.model1,
                                  self.model2)

    def load_to_gpu(self):
        self.model1.to_gpu()
        self.model2.to_gpu()
        self.config['has_cuda'] = True

    def load_to_cpu(self):
        self.model1.to_cpu()
        self.model2.to_cpu()
        self.config['has_cuda'] = False

    def load_model(self, model_path):
        model = load_model(model_path)
        # 设置参数
        self.config = model.config
        # 词编码器
        self.tokenizer = model.tokenizer
        # schema特征提取器
        self.schema_featurizer = model.schema_featurizer
        # column特征提取器
        self.column_featurizer = model.column_featurizer
        # 加载数据表
        self.schema = model.schema
        # 选择模型, 加载模型
        self.model1 = model.model1
        self.model2 = model.model2
        # 数据加载器
        self.schema_data = model.schema_data
        self.column_data = model.column_data
        # 预测器
        self.prediction = model.prediction
        self.config['has_cuda'] = False

    def dump_model(self, model_path):
        dump_model(self, model_path)
        print('Client model saved in path {0}'.format(model_path))

    def predict(self, question, first_sql=1, sec_sql=1, qid=0, size=50):
        return self.prediction.gen_json(question, first_sql, sec_sql, qid, size)



import numpy as np
import time


class BaseModel(object):
    """Define common interfaces for Hybrid Net models"""
    def to_cpu(self):
        raise NotImplementedError()

    def to_gpu(self):
        raise NotImplementedError()

    def train_on_batch(self, batch):
        raise NotImplementedError()

    def model_inference(self, model_inputs):
        """model prediction on processed features"""
        raise NotImplementedError()

    def save(self, model_path, name):
        raise NotImplementedError()

    def load(self, model_path, name):
        raise NotImplementedError()

    def dataset_inference(self, dataset):
        """model prediction on dataset"""
        print("model prediction start")
        start_time = time.time()
        # 获取预测输出
        model_outputs = self.model_inference(dataset.model_inputs)

        # 将输出按是否属于同一条数据进行分组
        # [{}, {}] 每一个dict都是一条数据对应的多条<>column<>question<>的输出汇总
        '''
            final_outputs = [                
                {
                    'column_func': [[1,1,1], [1,1,1], ..., []]
                    'sel': [[],[],...,[]]
                    'sel_num':[1, 1, 1, ..., 1]
                    'agg': [[], [],..., []] 
                    'where_num': [1, 1, 1, ..., 1]
                    'op': [[], [],..., []]
                    'value_num':[1, 1, 1, 1]
                    'value_start': [[], [], ..., []]
                    'value_end': [[], [], ..., []]
                    'relationship': [1, 1, 1, 1, 1]
                    'loss': [num1, num2, ..., ]
                },
                ...
            ]
        '''
        # 分组打包
        final_outputs = self.data_packle(dataset, model_outputs)
        print("model prediction end, time elapse: {0}".format(time.time() - start_time))
        assert len(dataset.input_features) == len(final_outputs)

        return final_outputs

    # 将数据按是否属于同一个问题进行打包
    def data_packle(self, dataset, model_outputs):
        final_outputs = []
        for pos in dataset.pos:
            final_output = {}
            # 分组
            for k in model_outputs:
                final_output[k] = model_outputs[k][pos[0]:pos[1], :]
            final_outputs.append(final_output)
        return final_outputs

    def parse_output_with_threshold(self, input_feature, model_output, sql_seq=0, threshold=0.9):
        '''
            解析预测输出, agg, select, where, conditions
            model_output: 是以input_features数据作为输入，经过模型预测的输出
            input_feature: 数据的特征，可以认为是label
            sql_seq: SQL顺序，用哪一个解码器去进行解析
            threshold: 阈值

            一个条验证语句对应的所有<>column<>question<>作为输入，输出是其结果的集合
            假设一条验证数据一共有三条<>column<>question<>
            输出：
            sel_num = 2
            where_num = 2
            relationship = [0, 0, 0]
            sel = [col_index1, col_index2]
            agg = [(col_index1, col_agg1), (col_index2,col_agg2)]
            where = [col_index1, col_index2]
            value_num = [0, 1]
            conditions = {
                            col_index1:(col_index1, op_index1, [(s1, e1)])
                            col_index2:(col_index2, op_index2, [(s1, e1),(s2, e2)])
                        }
        '''

        # 根据seq，选择解析器，选择概率大于阈值的列
        if sql_seq == 0:
            sel_relevant_prob = np.exp(model_output["sql_first_select_sim"][:, 0])
            sel_id_prob_exp = sorted(enumerate(sel_relevant_prob), key=lambda x: x[1], reverse=True)
            sel = []
            for col_index, sel_col_pro in sel_id_prob_exp:
                if sel_col_pro >= threshold:
                    sel.append(col_index)

            where_relevant_prob = np.exp(model_output["sql_first_where_sim"][:, 0])
            where_id_prob_exp = sorted(enumerate(where_relevant_prob), key=lambda x: x[1], reverse=True)
            where = []
            for col_index, value in where_id_prob_exp:
                if value >= threshold:
                    where.append(col_index)

        else:
            sel_relevant_prob = np.exp(model_output["sql_sec_select_sim"][:, 0])
            sel_id_prob_exp = sorted(enumerate(sel_relevant_prob), key=lambda x: x[1], reverse=True)
            sel = []
            # 后续可以加入概率
            for col_index, sel_col_pro in sel_id_prob_exp:
                if sel_col_pro >= threshold:
                    sel.append(col_index)
            sel_num = len(sel)

            where_relevant_prob = np.exp(model_output["sql_sec_where_sim"][:, 0])
            where_id_prob_exp = sorted(enumerate(where_relevant_prob), key=lambda x: x[1], reverse=True)
            where = []
            for col_index, value in where_id_prob_exp:
                if value >= threshold:
                    where.append(col_index)
            where_num = len(where)

        # aggregation
        agg = [(index, np.argmax(model_output["agg"][index, :])) for index in sel]  # 选择该列对应的概率最高的集合操作符agg

        # value_num
        value_num = [(index, np.argmax(model_output["value_num"][index, :])) for index in where]

        # relationship
        relationship = [(index, np.argmax(model_output["relationship"][index, :])) for index in set(sel + where)]

        # 构建条件 conditions(col_id, op, value_span[0], value_span[1])
        conditions = {}
        for index in range(0, len(where)):
            where_col_idx = where[index]
            span_list = self.get_span(where_col_idx, value_num[index][1], input_feature, model_output)  # 获取span
            op = np.argmax(model_output["op"][where_col_idx, :])  # 获取与列对应概率最大的op
            conditions[where_col_idx] = (where_col_idx, op, span_list)  # 构造conditions

        return sel, agg, where, value_num, conditions, relationship

    def get_span(self, col_index, vn, input_feature, model_output):
        """
            function: 获取第i列对应个值
            输入：
                col_index: 列id
                vn: 列对应值得数量
                input_feature: 输入特征
                model_output: 输入特征经过模型处理后的输出
            输出：
                span_list = [(start1, end1), (start2, end2)]
        """
        offset = 0  # 记录
        segment_ids = np.array(input_feature.segment_ids[col_index])
        # 获取question token 开始位置
        for j in range(len(segment_ids)):
            if segment_ids[j] == 1:
                offset = j
                break

        # 获取question token位置的value_start 和 value_end 的概率， 一共max_value_num对start, end
        value_start = model_output["value_start"][col_index, segment_ids == 1]
        value_end = model_output["value_end"][col_index, segment_ids == 1]
        l = len(value_start)  # question token序列长度

        '''
            求对数空间下的start和end的联合概率矩阵 log(P(start, end))=logP(start)P(end)=logP(start)+logP(end)
            转变为shape 是 (l, l) 矩阵, 矩阵形式
                    i am here <SEP>
            i       1  1  1    1
            am      1  1  1    1
            here    1  1  1    1
            <SEP>   1  1  1    1
        '''
        # 默认第0个value_pointer 不用作计算
        span_list = []
        for index in range(1, vn + 1):
            sum_mat = value_start[:, index].reshape((l, 1)) + value_end[:, index].reshape((1, l))
            # 对矩阵中的每一个单元[(i,j),value]按value进行降序排序
            span = (0, 0)
            for cur_span, _ in sorted(np.ndenumerate(sum_mat), key=lambda x: x[1], reverse=True):
                # cur_span shape 是 (i, j), 因为已经进行了排序，所以value就不重要了
                # 一个假设，start<=end, 只有矩阵 上三角+对角线 才是合理的取值区间
                # or cur_span[0] == l - 1 or cur_span[1] == l - 1 l-1 是SEP，不取
                if cur_span[1] < cur_span[0]:
                    continue
                span = cur_span
                span_list.append((span[0] + offset, span[1] + offset))
                break
        return span_list



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
from Text2JSON.train import utils


class ProcessedSchema(object):
    def __init__(self):
        self.url_id = ''
        self.url = ''
        self.method = ''
        self.url_note = ''
        self.header = []
        self.types = []
        self.space = []
        self.notes = []

    def to_dict(self):
        return {
            "id": self.url_id,
            "url": self.url,
            "method": self.method,
            "url_note": self.url_note,
            "header": self.header,
            "types": self.types,
            "space": self.space,
            "notes": self.notes
        }


def gen_processed_schema(source, target):
    schema_list = []
    for data in utils.load_jsonl(source):
        schema = ProcessedSchema()
        schema.url_id = data['id']
        schema.url = data['url']
        schema.method = data['method']
        schema.url_note = data['url_note']
        schema.header = ["order&asc", "order&desc"]
        schema.types = ["string", "string"]
        schema.space = ["closed", "closed"]
        schema.notes = ["升序,最少,最小,最低", "降序,最大,最高"]
        params = data['params']
        for param in params:
            param_name = param['param_name']
            param_note = param['param_note']
            value_type = param['value_type']
            value_space = param['value_space']
            value_range = param['value_range']
            value_note = param['value_note']
            if param_name == 'order':
                continue
            if value_space == 'closed':
                for index, value in enumerate(value_range):
                    schema.header.append(param_name+'&'+value)
                    schema.types.append(value_type)
                    schema.space.append(value_space)
                    if param_note == '':
                        schema.notes.append(value_note[index])
                    else:
                        schema.notes.append(param_note+','+value_note[index])
            else:
                schema.header.append(param_name)
                schema.types.append(value_type)
                schema.space.append(value_space)
                schema.notes.append(param_note)
        schema_list.append(schema)
    schema_jsonl = [schema.to_dict() for schema in schema_list]
    utils.dump_jsonl(schema_jsonl, target)



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import os
import datetime
import torch.utils.data as torch_data
from Text2JSON.train import utils
from Text2JSON.train.data_utils import *
from Text2JSON.evaluator.schema_evaluator import SchemaEvaluator
from Text2JSON.featurizer.schema_parser import schema_parsed
from Text2JSON.model.schema_model import HydraSchemaTorch
from Text2JSON.featurizer.schema_featurizer import HydraSchemaFeaturizer
from Text2JSON.evaluator.column_evaluator import ColumnEvaluator
from Text2JSON.model.column_model import HydraColumnTorch
from Text2JSON.predict.prediction import Predict


class ModelTrain(object):
    def __init__(self, conf_path, output_path, train_data_path, dev_data_path, schema_path, bert_path):
        """
            param: conf_path 配置文件路径
            param: output_path 模型输出路径
        """
        # 记载模型路径
        self.config = utils.read_conf(conf_path)
        # 训练数据存放位置
        self.config['train_data_path'] = train_data_path
        # 验证数据存放位置
        self.config['dev_data_path'] = dev_data_path
        # URL 处理后存放的路径
        self.config['schema_path'] = schema_path
        # 语言预训练模型存放位置
        self.config['bert_path'] = bert_path
        # 模型输出位置
        self.output_path = output_path
        # 加载tokenizer
        self.tokenizer = utils.create_tokenizer(self.config)
        # 加载数据表
        self.schema = schema_parsed()
        self.schema.load_schema(self.config, self.tokenizer)
        # 加载特征解析器
        self.schema_featurizer = HydraSchemaFeaturizer(self.config, self.tokenizer)
        # 加载特征解析器
        self.column_featurizer = HydraColumnFeaturizer(self.config, self.tokenizer)
        # 设置GPU还是CPU版本
        if 'CUDA_VISIBLE_DEVICES' in os.environ.keys():
            self.config['has_cuda'] = True
        else:
            self.config['has_cuda'] = False

    def train_column(self):
        # 记载训练数据，并将训练数据封装到dataloader中
        train_data = load_train_column_data(self.config, self.schema, self.column_featurizer, True, True, False,
                                            note=True)
        train_data_loader = torch_data.DataLoader(train_data,
                                                  batch_size=int(self.config["column_batch_size"]),
                                                  shuffle=True,
                                                  pin_memory=True)

        # negative train data
        negative_train_data = load_train_column_data(self.config, self.schema, self.column_featurizer, True, False,
                                                     True, note=True)
        negative_train_data_loader = torch_data.DataLoader(negative_train_data,
                                                           batch_size=int(self.config["neg_column_train_batch_size"]),
                                                           shuffle=True,
                                                           pin_memory=True)
        # 加载验证数据， 分别是eval column, eval table
        eval_data = load_eval_column_data(self.config, self.schema, self.column_featurizer, True, True, negative=False,
                                          note=True)

        # 加载模型
        model = HydraColumnTorch(self.config)
        if self.config['has_cuda']:
            model.to_gpu()
        # 加载验证器
        evaluator = ColumnEvaluator(self.output_path, self.config, model, eval_data)
        # 设置训练步数
        num_samples = len(train_data)  # 训练样本总数量
        self.config["column_num_train_steps"] = int(
            num_samples * int(self.config["column_epochs"]) / int(self.config["column_batch_size"]))  # 设置训练步数
        print("total_steps: {0}, warm_up_steps: {1}".format(self.config["column_num_train_steps"],
                                                            self.config["column_num_warmup_steps"]))

        # 初始化参数和优化器
        model.init_optimizer()

        # 列训练
        loss_avg, step, epoch, best_overall, result_str = 0.0, 0, 0, -0.1, None
        print("Column training")
        while True:
            for batch_id, batch in enumerate(train_data_loader):
                cur_loss = model.train_on_batch(batch)
                for neg_batch_id, neg_batch in enumerate(negative_train_data_loader):
                    if neg_batch_id >= int(self.config["neg_column_train_epoch"]):
                        break
                    cur_loss += model.train_on_batch(neg_batch)
                loss_avg = (loss_avg * step + cur_loss) / (step + 1)
                step += 1
                if batch_id % 100 == 0:
                    currentDT = datetime.datetime.now()
                    print("[{3}] epoch {0}, batch {1}, column_batch_loss={2:.4f}".format(epoch, batch_id, cur_loss,
                                                                                         currentDT.strftime(
                                                                                             "%m-%d %H:%M:%S")))

            r_str, acc = evaluator.eval(epoch)
            if acc['overall'] > best_overall:
                model.save(self.output_path, 'column')
                best_overall = acc['overall']
                result_str = r_str
            epoch += 1
            if epoch >= int(self.config["column_epochs"]):
                break
        # 将最好结果写入
        evaluator.append('final result', result_str)
        # 删除模型
        model.to_cpu()
        del model

    def train_schema(self):
        # 加载正样本train table data
        train_schema_data = load_train_schema_data(self.config, self.schema, self.schema_featurizer, True, True,
                                                   negative=False, note=True)
        train_schema_data_loader = torch_data.DataLoader(train_schema_data,
                                                         batch_size=int(self.config["schema_batch_size"]),
                                                         shuffle=True,
                                                         pin_memory=True)
        # 加载负样本train table data
        negative_train_schema_data = load_train_schema_data(self.config, self.schema, self.schema_featurizer, True,
                                                            False, True, note=True)
        negative_train_schema_data_loader = torch_data.DataLoader(negative_train_schema_data,
                                                                  batch_size=int(
                                                                      self.config["neg_schema_train_batch_size"]),
                                                                  shuffle=True,
                                                                  pin_memory=True)
        # 加载验证数据， 分别是eval column
        eval_schema_data = load_dev_schema_data(self.config, self.schema, self.schema_featurizer, True, True, True,
                                                note=True)
        # 加载模型
        model = HydraSchemaTorch(self.config)
        if self.config['has_cuda']:
            model.to_gpu()
        # 加载验证器
        evaluator = SchemaEvaluator(self.output_path, self.config, model, eval_schema_data)

        num_samples = len(train_schema_data_loader)  # 训练样本总数量
        self.config["schema_num_train_steps"] = int(
            num_samples * int(self.config["schema_epochs"]) / int(self.config["schema_batch_size"]))  # 设置训练步数
        print("start training")
        print("total_steps: {0}, warm_up_steps: {1}".format(self.config["schema_num_train_steps"],
                                                            self.config["schema_num_warmup_steps"]))

        # 初始化参数和优化器
        model.init_optimizer()

        # 表训练
        loss_avg, step, epoch, best_overall, result_str = 0.0, 0, 0, -0.1, None
        print("Schema training")
        while True:
            for batch_id, batch in enumerate(train_schema_data_loader):
                cur_loss = model.train_on_batch(batch)
                for neg_batch_id, neg_batch in enumerate(negative_train_schema_data_loader):
                    if neg_batch_id >= int(self.config["neg_schema_train_epochs"]):
                        break
                    cur_loss += model.train_on_batch(neg_batch)
                    loss_avg = (loss_avg * step + cur_loss) / (step + 1)
                if batch_id % 10 == 0:
                    currentDT = datetime.datetime.now()
                    print("[{3}] epoch {0}, batch {1}, table_batch_loss={2:.4f}".format(epoch, batch_id, cur_loss,
                                                                                        currentDT.strftime(
                                                                                            "%m-%d %H:%M:%S")))
            r_str, acc = evaluator.eval(epoch)
            if acc['overall'] > best_overall:
                model.save(self.output_path, 'schema')
                best_overall = acc['overall']
                result_str = r_str
            epoch += 1
            if epoch >= int(self.config["schema_epochs"]):
                break
        # 将最好结果写入
        evaluator.append('final result', result_str)
        # 删除模型
        model.to_cpu()
        del model

    def train(self):
        """
            function: 训练模型
        """
        self.train_schema()
        print('Schema Model train done')
        self.train_column()
        print('Column Model train done')

    def dump_client_model(self):
        """
            function: 导出客户端模型
        """
        schema_model = HydraSchemaTorch(self.config)
        schema_model.load(self.output_path, 'schema')
        column_model = HydraColumnTorch(self.config)
        column_model.load(self.output_path, 'column')
        predict = Predict()
        predict.set_params(self.config, self.tokenizer, self.schema_featurizer,
                           self.column_featurizer, self.schema, schema_model, column_model)
        predict.dump_model(os.path.join(self.output_path, "model.pt"))

        # 删除子模型
        os.remove(os.path.join(self.output_path, "model_schema.pt"))
        os.remove(os.path.join(self.output_path, "model_column.pt"))



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import os
from Text2JSON.featurizer.column_featurizer import ColumnDataset
from Text2JSON.model.base_model import BaseModel


class ColumnEvaluator:
    def __init__(self, output_path, config, model: BaseModel, eval_data):
        """
            评估网络效果
            output_path: 结果输出路径
            config: 参数配置
            eval_data: 需要验证的数据
            hydra_featurizer: 特征提取器
            model: 模型
            note: 注释
        """
        self.config = config
        self.model = model
        self.eval_history_file = os.path.join(output_path, "column_eval.log")
        f = open(self.eval_history_file, 'w')
        f.close()
        # 预测结果存放位置
        self.eval_data = eval_data

    def eval_column(self, eval_data: ColumnDataset):
        # 10项评价指标
        items = ['overall', 'sc', 'agg', 'wc', 'op', 'vn', 'val', 'rel']

        # 建立映射
        acc = {k: 0.0 for k in items}
        cnt = 0
        '''
            每一个dict都是一条数据对应的多条<>column<>question<>的输出汇总
            model_outputs = [                
                {
                    'column_func': [[1,1,1], [1,1,1], ..., []]
                    'seq': [1, 1, 1, 0]
                    'sel': [[],[],...,[]]
                    'sel_num':[1, 1, 1, ..., 1]
                    'agg': [[], [],..., []] 
                    'where_num': [1, 1, 1, ..., 1]
                    'op': [[], [],..., []]
                    'value_num':[1, 1, 1, 1]
                    'value_start': [[], [], ..., []]
                    'value_end': [[], [], ..., []]
                    'relationship': [1, 1, 1, 1, 1]
                    'loss': [num1, num2, ..., ]
                },
                ...
            ]
        '''
        # 一次给整个数据集
        model_outputs = self.model.dataset_inference(eval_data)
        # input_features 作为label, model_outputs 作为predict
        for input_feature, model_output in zip(eval_data.input_features, model_outputs):
            # 设置当前验证数据的七项指标
            cur_acc = {k: 1 for k in acc.keys() if k != "overall"}

            sel_label = input_feature.sel
            agg_label = input_feature.agg
            wc_label = input_feature.where  # 获取where_clause对应的列
            op_label = input_feature.op
            value_num_label = input_feature.value_num
            value_start_end_label = []
            for index, value in enumerate(input_feature.where):
                if value == 0:
                    continue
                value_num = input_feature.value_num[index]
                value_start_end_label.append([(input_feature.value_start[index][i],
                                               input_feature.value_end[index][i])
                                              for i in range(0, value_num)])
            relationship_label = input_feature.relationship
            '''
                输出：
                seq =  [0, 0, 0]
                relationship = [0, 0, 0]
                sel_num = 2
                sel = [col_index1, col_index2]
                agg = [col_agg1, col_agg2]
                where_num = 2
                where = [col_index1, col_index2]
                value_num = [0, 1]
                conditions = {
                                col_index1:(col_index1, op_index1, [(s1, e1)])
                                col_index2:(col_index2, op_index2, [(s1, e1),(s2, e2)])
                            }
            '''
            # 选择解码方式
            # print('seq', input_feature.seq)
            if 1 not in input_feature.seq:
                sql_seq = 0
            else:
                sql_seq = 1

            sel, agg, where, value_num, conditions, relationship = \
                self.model.parse_output_with_threshold(input_feature, model_output, sql_seq,
                                                       float(self.config['column_threshold']))

            # relationship evaluate
            for pre_col_index, value in relationship:
                if int(relationship_label[pre_col_index]) != value:
                    cur_acc['rel'] = 0
                    break

            # (sel_col_index1, sel_col_index2)
            # (agg1, agg2)
            # sc evaluate
            for index, value in enumerate(sel_label):
                if value == 1 and index not in sel:
                    cur_acc['sc'] = 0
                    break

            # agg evaluate
            agg_dict = {}
            for index, value in agg:
                agg_dict[index] = value
            for index, value in enumerate(agg_label):
                if value == 0:
                    continue
                if index not in agg_dict.keys():
                    cur_acc['agg'] = 0
                    break
                if value != agg_dict[index]:
                    cur_acc['agg'] = 0
                    break

            # wc evaluate
            for index, value in enumerate(wc_label):
                if value == 1 and index not in where:
                    cur_acc['wc'] = 0
                    break

            # vn evaluate
            value_num_dict = {}
            for index, value in value_num:
                value_num_dict[index] = value

            for index, value in enumerate(value_num_label):
                if value == 0:
                    continue
                if index not in value_num_dict.keys():
                    cur_acc['vn'] = 0
                    break
                if value != value_num_dict[index]:
                    cur_acc['vn'] = 0
                    break

            for w in where:
                _, op, span_list = conditions[w]
                if op != op_label[w]:
                    cur_acc["op"] = 0

                for index, span in enumerate(span_list):
                    sv = span[0]
                    se = span[1]
                    if sv != input_feature.value_start[w][index] or \
                            se != input_feature.value_end[w][index]:
                        cur_acc["val"] = 0

            # 加到中的acc当中
            for k in cur_acc:
                acc[k] += cur_acc[k]

            # 计算overall
            all_correct = 0 if 0 in cur_acc.values() else 1
            cur_acc['overall'] = all_correct
            acc["overall"] += all_correct

            # 样本数自增1
            cnt += 1

        result_str = []
        for item in items:
            result_str.append(item + ":{0:.1f}".format(acc[item] * 100.0 / cnt))

        result_str = ", ".join(result_str)

        return result_str, acc

    def eval(self, epochs):
        print('column eval_file', self.config['dev_data_path'])
        result_str, acc = self.eval_column(self.eval_data)
        print("[ epochs " + str(epochs) + "] " + result_str)
        with open(self.eval_history_file, "a+", encoding="utf8") as f:
            f.write("[{0}, epoch {1}] ".format(self.config['dev_data_path'], epochs) + result_str + "\n")
        return result_str, acc

    # 将最好结果写入
    def append(self, state, result_str):
        with open(self.eval_history_file, "a+", encoding="utf8") as f:
            f.write("[{0}, {1}] ".format(self.config['dev_data_path'], state) + result_str + "\n")




import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
from Text2JSON.train.utils import load_jsonl_by_key, load_jsonl, dump_jsonl
from Text2JSON.entity_named_recog.entity_recognition import entity_recognition


class ProcessedData:
    def __init__(self):
        self.qid = ''
        self.table_id = []
        self.question = ''
        self.tokens = []
        self.char_to_word = []
        self.word_to_char_start = []
        self.sql = {'clause': [], 'relationship': '0'}
        self.valid = ''

    def to_dict(self):
        return {
            'qid': self.qid,
            'table_id': self.table_id,
            'question': self.question,
            'tokens': self.tokens,
            'char_to_word': self.char_to_word,
            'word_to_char_start': self.word_to_char_start,
            'sql': self.sql,
            'valid': self.valid
        }


# 记录WHERE Clause 条件
class Condition:
    def __init__(self, col, op, value):
        self.col = col
        self.op = op
        self.value = value
        self.value_list = value.replace('[', '').replace(']', '').replace('"', '').split(',')

    def parse_with_table(self, op_map, table):
        return [[table['header'].index(self.col), op_map[self.op], len(self.value_list), self.value_list]]


def gen_processed_data(source, target, table_file):
    # 记录操作符映射规则
    # cond_ops = ['=', '!=', '>', '<', '>=', '<=', 'like', 'isnull', 'notnull', 'OP']
    op_map = {'': 0, 'eq': 1, 'not': 2, 'gt': 3, 'lt': 4, 'gte': 5, 'lte': 6, 'like': 7,
              'isnull': 8, 'notnull': 9, 'btw': 10, 'in': 11}
    agg_map = {'': 0, 'eq': 1, 'not': 2, 'gt': 3, 'lt': 4, 'gte': 5, 'lte': 6, 'like': 7,
               'isnull': 8, 'notnull': 9, 'btw': 10, 'in': 11}
    tables = load_jsonl_by_key('url', table_file)
    train_sql_list = []
    for line in load_jsonl(source):
        qid = line['qid']
        question = line['question'].replace(' ', '').strip()
        question = entity_recognition(question)
        tokens = [char for char in question]
        question_tokens = ' '.join(tokens)+' '
        train = ProcessedData()
        train.tokens = tokens
        train.qid = qid

        # do char_to_word
        char_to_word = []
        span = 0
        for i in question_tokens:
            char_to_word.append(span)
            if i == ' ':
                span += 1
        train.char_to_word = char_to_word

        # do word_to_char_start
        space_list = [i for i, x in enumerate(question_tokens) if x == ' ']
        word_to_char_start = [0]
        word_to_char_start.extend([space_list[index] + 1 for index in range(0, len(space_list) - 1)])
        train.word_to_char_start = word_to_char_start

        # 完整query语句
        train.sql['relationship'] = 1 if len(line['query']) == 2 else 0
        try:
            for k in range(0, len(line['query'])):
                temp_query = line['query'][k]
                # 获取表名
                table_name = temp_query['url']
                # 获取表头
                header = [value for value in tables[table_name]['header']]

                # 设置table数据
                train.table_id.append(tables[table_name]['id'])

                # select_clause 所涉及到的列, 默认加上一个占位符, 如果后续不在
                select_list = []
                # select_clause 所涉及到的列集合运算符
                agg_list = []
                # where_clause 所涉及到的列，操作符，值
                cond_list = []

                if 'params' not in temp_query.keys():
                    '不用加占位符'
                else:
                    for l in range(0, len(temp_query['params'])):
                        temp_param = temp_query['params'][l]
                        col = temp_param['name']
                        option = temp_param['option']
                        value = temp_param['value']
                        value_list = value.replace('[', '').replace(']', '').replace('"', '').split(',')
                        flag = value_in_question(question, value_list)
                        if flag:
                            cond_list.append(Condition(col, option, value))
                        else:
                            for value in value_list:
                                select_list.append(col + '&' + value)
                                agg_list.append(option)

                # 解析sql语句,并且设置train数据
                value_start_end = {}
                sel = [header.index(sel_col) for sel_col in select_list]
                agg = [agg_map[agg] for agg in agg_list]
                conditions = []
                for con in cond_list:
                    conditions.extend(con.parse_with_table(op_map, tables[table_name]))
                for condition in conditions:
                    value_list = condition[3]
                    for value in value_list:
                        value_start_end[value] = find_token_start_end(question_tokens, question, value)
                train.sql['clause'].append({'seq': k, 'sel': sel, 'agg': agg,
                                            'conds': conditions,
                                            'value_start_end': value_start_end})

            # # 设置train数据
            train.question = ' '.join(tokens)+' '
            train.valid = True
            train_sql_list.append(train)
        except ValueError:
            print('数据解析错误, qid={0}的标注数据无法识别'.format(qid))

    # 导出到文件中
    dump_jsonl([sql.to_dict() for sql in train_sql_list], target)


def value_in_question(question, value_list):
    flag = True
    for value in value_list:
        if value in question:
            continue
        else:
            flag = False
            break
    return flag


# 找出token的开始位置和结束位置
def find_token_start_end(question, question_without_token, value):
    if value not in question_without_token:
        return [-1, -1]

    question_list = question.split(' ')
    candidate_token = []
    for index, token in enumerate(question_list):
        if token in value and token != '':
            candidate_token.append([token, index])

    token_list = []
    for pair in candidate_token:
        token = pair[0]
        index = pair[1]
        token_list.append(token)
        sub_token = ''.join(token_list)
        if len(sub_token) < len(value):
            continue
        elif len(sub_token) > len(value):
            length = len(sub_token)
            while length > len(value):
                token_list.pop(0)
                sub_token = ''.join(token_list)
                length = len(sub_token)
        if sub_token == value:
            start_token_index = index - len(token_list) + 1
            # end_token_index = index + 1
            end_token_index = index
            return [start_token_index, end_token_index]

    print('error, 逻辑错误',value,question_without_token)
    return [-1, -1]



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import torch
import os
import numpy as np
import transformers
from Text2JSON.train import utils
from Text2JSON.model.base_model import BaseModel
from torch import nn


class HydraSchemaTorch(BaseModel):
    def __init__(self, config):
        # 配置文件config
        self.config = config
        # 根据config设置，加载对应的HydraNet
        self.schema_model = HydraSchemaNet(config)
        # 定义优化器，学习率优化器
        self.optimizer, self.scheduler = None, None

    def init_optimizer(self):
        # 衰退设定
        no_decay = ["bias", "LayerNorm.weight"]
        # 返回model的所有参数的(name, tensor)的键值对, 并且设置其值
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in self.schema_model.named_parameters() if not any(nd in n for nd in no_decay)],
                "weight_decay": float(self.config["decay"]),
            },
            {"params": [p for n, p in self.schema_model.named_parameters() if any(nd in n for nd in no_decay)],
             "weight_decay": 0.0},
        ]
        # AdamW是实现了权重衰减的优化器
        self.optimizer = transformers.AdamW(optimizer_grouped_parameters,
                                            lr=float(self.config["schema_learning_rate"]))
        # 设置耐心系数
        self.scheduler = transformers.get_cosine_schedule_with_warmup(
            self.optimizer,
            num_warmup_steps=int(self.config["schema_num_warmup_steps"]),
            num_training_steps=int(self.config["schema_num_train_steps"]))

        # 梯度清零
        self.optimizer.zero_grad()

    def train_on_batch(self, batch):
        self.schema_model.train()
        # 加载数据到cuda
        if self.config['has_cuda']:
            for k, v in batch.items():
                batch[k] = v.to(torch.device("cuda:0"))

        # 默认调用model.forward函数
        batch_loss = torch.mean(self.schema_model(**batch)["loss"])  # 计算loss

        batch_loss.backward()  # loss反向传播
        torch.nn.utils.clip_grad_norm_(self.schema_model.parameters(), 1.0)  # 用L2范数作为正则
        self.optimizer.step()  # 优化
        self.scheduler.step()  # 更新学习率
        self.optimizer.zero_grad()  # 梯度清零

        if self.config['has_cuda']:
            # 返回一定需要detach, 否则会累积梯度
            return batch_loss.cpu().detach().numpy()
        else:
            return batch_loss.detach().numpy()

    # 模型预测
    def model_inference(self, model_inputs):
        # 不启用 BatchNormalization 和 Dropout
        self.schema_model.eval()
        model_outputs = {}

        # batch_size = 512  # 设置步长， 这一步有可能出现内存溢出，内存太大了
        batch_size = int(self.config['schema_evaluator_batch_size'])
        # 在(0~shape[0])之间遍历，步长为batch_size
        '''
            一个batch的输入格式
            input_tensor={
                            'input_ids': [],
                            'input_mask': [],
                            'segment_ids': []
                            }
        '''
        for start_idx in range(0, model_inputs["input_ids"].shape[0], batch_size):
            if self.config['has_cuda']:
                input_tensor = {
                    # 将batch中的数据分组，以batch_size为单位，分组
                    # 只取输入编码特征，不取其他特征
                    k: torch.from_numpy(model_inputs[k][start_idx:start_idx + batch_size]).to(torch.device("cuda:0"))
                    for k in ["input_ids", "input_mask", "segment_ids"]
                }
            else:
                input_tensor = {
                    # 将batch中的数据分组，以batch_size为单位，分组
                    # 只取输入编码特征，不取其他特征
                    k: torch.from_numpy(model_inputs[k][start_idx:start_idx + batch_size])
                    for k in ["input_ids", "input_mask", "segment_ids"]
                }

            # 因为只做预测，所以不更新梯度
            with torch.no_grad():
                # 数据输入模型
                model_output = self.schema_model(**input_tensor)

            # 将输出按关键字key进行分组
            for k, out_tensor in model_output.items():
                if out_tensor is None:
                    continue
                if k not in model_outputs:
                    model_outputs[k] = []
                if self.config['has_cuda']:
                    # 将数据转移到CPU，否则就会导致GPU内存溢出
                    model_outputs[k].append(out_tensor.cpu().detach().numpy())
                else:
                    model_outputs[k].append(out_tensor.detach().numpy())

        # 将多个batch结果合在一起
        for k in model_outputs:
            model_outputs[k] = np.concatenate(model_outputs[k], 0)

        # 写一下注释
        return model_outputs

    def save(self, model_path, name):
        save_path = os.path.join(model_path, "model_{0}.pt".format(name))
        torch.save(self.schema_model.state_dict(), save_path)
        print("Schema Model saved in path: %s" % save_path)

    def load(self, model_path, name):
        pt_path = os.path.join(model_path, "model_{0}.pt".format(name))
        self.schema_model.load_state_dict(torch.load(pt_path))
        print("Schema Model loaded from {0}".format(pt_path))

    def to_cpu(self):
        self.schema_model.cpu()

    def to_gpu(self):
        self.schema_model.to(torch.device("cuda:0"))


class HydraSchemaNet(nn.Module):
    def __init__(self, config):
        super(HydraSchemaNet, self).__init__()
        self.config = config

        # load bert model
        self.base_model = utils.create_base_model(config)

        drop_rate = float(config["drop_rate"]) if "drop_rate" in config else 0.0
        self.dropout = nn.Dropout(drop_rate)

        bert_hid_size = self.base_model.config.hidden_size  # 获取隐变量维度

        # 计算SQL1 的 table 相似度
        self.schema_first_sim = nn.Linear(bert_hid_size, 2)
        # 计算SQL2 的 table 相似度
        self.schema_sec_sim = nn.Linear(bert_hid_size, 2)

    def forward(self, input_ids, input_mask, segment_ids, schema_rel=None, schema_mask=None):
        # print('table_rel', table_rel)
        # print('table_seq', table_seq)
        # print('table_mask', table_mask)
        # 获取输出
        # bert_output 是最后一层编码层的特征向量 <cls>h,h,...,h<sep>q,q,...,q<sep>
        # pooled_output 是 cls
        _, pooled_output = self.base_model(
            input_ids=input_ids,
            attention_mask=input_mask,
            token_type_ids=segment_ids)

        # dropout 放弃部分值
        pooled_output = self.dropout(pooled_output)

        # schema simi
        schema_first_logit = self.schema_first_sim(pooled_output)
        schema_sec_logit = self.schema_sec_sim(pooled_output)

        loss = None
        # Not None 对应train的情况
        # None 对应inference的情况
        if schema_rel is not None:
            schema_rel = schema_rel.type_as(schema_first_logit)
            bceloss = nn.BCEWithLogitsLoss(reduction="none")  # 将sigmod函数和Binary CrossEntropyLoss合成一步

            # 计算table similarity loss
            loss = bceloss(schema_first_logit[:, 0], schema_rel) * (1 - schema_mask)
            loss += bceloss(schema_sec_logit[:, 0], schema_rel) * schema_mask

        log_sigmoid = nn.LogSigmoid()
        return {"schema_first_sim": log_sigmoid(schema_first_logit),
                'schema_sec_sim': log_sigmoid(schema_sec_logit),
                "loss": loss}



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
from Text2JSON.entity_named_recog.minute_entity_recognition import minute_recon
from Text2JSON.entity_named_recog.hour_entity_recognition import hour_recon
from Text2JSON.entity_named_recog.day_entity_recognition import day_recon
from Text2JSON.entity_named_recog.month_entity_recognition import month_recon
from Text2JSON.entity_named_recog.quarter_entity_recognition import quart_recon
from Text2JSON.entity_named_recog.year_entity_recognition import year_recon
from Text2JSON.entity_named_recog.proper_noun_recognition import proper_recon
import chinese2digits as ctd
import copy


def entity_recognition(original_line):
    line = copy.deepcopy(original_line)
    line = line.replace('：', ':')
    line = line.replace(' ', '')
    placeholders_list = {}
    line = proper_recon(line, placeholders_list)
    line = minute_recon(line, placeholders_list)
    line = hour_recon(line, placeholders_list)
    line = day_recon(line, placeholders_list)
    line = month_recon(line, placeholders_list)
    line = quart_recon(line, placeholders_list)
    line = year_recon(line, placeholders_list)
    line = ctd.takeChineseNumberFromString(line)['replacedText']
    for holder, data in placeholders_list.items():
        line = line.replace(holder, data[0])
    line = line.replace('““', '“')
    line = line.replace('””', '”')
    return line


if __name__ == '__main__':
    ''
    # example1 = "帮我查一下最近1个月内有哪些范本有新的评价"
    # original_line, line = entity_recognition(example1)
    # print(original_line)
    # print(line)
    # example2 = "我创建一个采购办公用品的合同，金额30万"
    # original_line, line = entity_recognition(example2)
    # print(original_line)
    # print(line)
    # example3 = "帮我用办公用品采购合同范本创建一个合同，金额30万，相对方是北京慧点科技有限公司"
    # original_line, line = entity_recognition(example3)
    # print(original_line)
    # print(line)
    # example4 = "帮我查一下我10月3日提交的“办公用品采购合同”现在到谁审批了"
    # original_line, line = entity_recognition(example4)
    # print(original_line)
    # print(line)
    # example5 = "我正在执行的合同里有哪些合同的相对方最近1个月发生过风险事件"
    # original_line, line = entity_recognition(example5)
    # print(original_line)
    # print(line)
    # example6 = "帮我查一下今年西安分公司销售合同签订的情况（份数、金额）"
    # original_line, line = entity_recognition(example6)
    # print(original_line)
    # print(line)
    # example7 = "今年发生解除的有哪些合同"
    # original_line, line = entity_recognition(example7)
    # print(original_line)
    # print(line)
    # example8 = "帮我导出一下今年合同的台账"
    # original_line, line = entity_recognition(example8)
    # print(original_line)
    # print(line)
    # example9 = "帮我导出本季度, 西安分公司合同的台账，包含合同名称、合同金额、相对方、合同编号"
    # original_line, line = entity_recognition(example9)
    # print(original_line)
    # print(line)
    # example10 = "今年一共发生了多少案件"
    # original_line, line = entity_recognition(example10)
    # print(original_line)
    # print(line)
    # example11 = "有多少案件今年已经结案"
    # original_line, line = entity_recognition(example11)
    # print(original_line)
    # print(line)
    # example12 = "帮我导出今年新发案件的台账"
    # original_line, line = entity_recognition(example12)
    # print(original_line)
    # print(line)
    # example13 = "帮我查一下最近1个月内，上个月和下一个月有哪些范本有新的评价"
    # original_line, line = entity_recognition(example13)
    # print(original_line)
    # print(line)
    # example14 = "帮我查一下本季度，上一季度，下一季度，最近一个季度，上季度，下季度"
    # original_line, line = entity_recognition(example14)
    # print(original_line)
    # print(line)
    # example15 = '15年第1季度，2014年第二季度，2021年第三季度，08年第四季度，1998年第二季度，本季度，上季度，下季度，第一季度，第二季度，第3季度，第4季度'
    # original_line, line = entity_recognition(example15)
    # print(original_line)
    # print(line)
    # example16 = '1月，二月，3月，16年2月，17年12月'
    # original_line, line = entity_recognition(example16)
    # print(original_line)
    # print(line)
    # example17 = '帮我订一个会议室，明天上午9：00 - 11: 30，9人参与，要求有投影仪'
    # original_line, line = entity_recognition(example17)
    # print(original_line)
    # print(line)
    # example18 = '凌晨三点，早上六点，上午八点，中午十二点，下午六点，晚上九点，傍晚五点'
    # original_line, line = entity_recognition(example18)
    # print(original_line)
    # print(line)
    # example19 = '凌晨3点，早上6点，上午8点，中午12点，下午6点，晚上9点，傍晚5点'
    # original_line, line = entity_recognition(example19)
    # print(original_line)
    # print(line)
    # example20 = '凌晨三点二十分，上午八点五十五分，中午十二点十分，下午两点五十五分，晚上九点二十分，傍晚五点半'
    # original_line, line = entity_recognition(example20)
    # print(original_line)
    # print(line)
    # example21 = '凌晨3点，早上6点，上午8点，中午12点，下午6点，晚上9点，傍晚5点'
    # original_line, line = entity_recognition(example21)
    # print(original_line)
    # print(line)
    # example22 = '上午六点五十五分, 早上八点十分， 早上八点零分， 早上八点二十分，早上八点零分'
    # original_line, line = entity_recognition(example22)
    # print(original_line)
    # print(line)
    # example23 = '下午六点五十五分, 傍晚八点十分， 晚上八点零分， 晚上八点十分， 晚上八点零分'
    # original_line, line = entity_recognition(example23)
    # print(original_line)
    # print(line)
    # example24 = '中午十二点五十五分, 正午十一点十分， 正午一点零分， 中午一点十分'
    # original_line, line = entity_recognition(example24)
    # print(original_line)
    # print(line)
    # example24 = '凌晨二点五十五分, 凌晨十一点十分， 凌晨一点零分， 凌晨一点十分'
    # original_line, line = entity_recognition(example24)
    # print(original_line)
    # print(line)
    # example25 = '二点五十五分, 十一点十分， 二十一点零分， 二十四点十分'
    # original_line, line = entity_recognition(example25)
    # print(original_line)
    # print(line)
    # example26 = '二点55分, 十一点10分， 二十一点0分， 二十四点20分'
    # original_line, line = entity_recognition(example26)
    # print(original_line)
    # print(line)
    # example26 = '11日二点55分, 十二日十一点10分， 下一天二十一点0分， 昨天二十四点20分'
    # original_line, line = entity_recognition(example26)
    # print(original_line)
    # print(line)
    # example27 = '十二日十一点10分， 下一天二十一点0分， 昨天二十四点20分'
    # original_line, line = entity_recognition(example27)
    # print(original_line)
    # print(line)
    # example27 = '11日3点55分 - 3点20分'
    # original_line, line = entity_recognition(example27)
    # print(original_line)
    # print(line)
    # example28 = '18年12月11日3点55-4点20'
    # original_line, line = entity_recognition(example28)
    # print(original_line)
    # print(line)
    # example29 = '下午3点五十分'
    # original_line, line = entity_recognition(example29)
    # print(original_line)
    # print(line)
    # example29 = '30分钟后，五分钟后，60分钟前，十分钟前'
    # original_line, line = entity_recognition(example29)
    # print(original_line)
    # print(line)
    # example29 = '5小时前，三十三小时前，14小时后，五小时后'
    # original_line, line = entity_recognition(example29)
    # print(original_line)
    # print(line)
    # example29 = '3天后，十六天后，5天前，十三天前'
    # original_line, line = entity_recognition(example29)
    # print(original_line)
    # print(line)
    #
    # example29 = '3天后下午6点-晚上9点'
    # original_line, line = entity_recognition(example29)
    # print(original_line)
    # print(line)
    #
    # example29 = '下个月6日下午5点30-晚上7点45'
    # original_line, line = entity_recognition(example29)
    # print(original_line)
    # print(line)




import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
from Text2JSON.train.utils import load_jsonl, dump_jsonl
from Text2JSON.entity_named_recog.entity_recognition import entity_recognition


class RawSchema(object):
    def __init__(self):
        self.id = None
        self.method = None
        self.url = None
        self.url_note = ''
        self.params = {}

    def to_dict(self):
        return {
            'id': self.id,
            'method': self.method,
            'url': self.url,
            'url_note': self.url_note,
            'params': [param.to_dict() for param in self.params.values()]
        }


class RawParam(object):
    def __init__(self):
        self.param_name = None
        self.param_note = ''
        self.value_type = None
        self.value_space = None
        self.value_range = []
        self.value_note = []

    def to_dict(self):
        return {
            'param_name': self.param_name,
            'param_note': self.param_note,
            'value_type': self.value_type,
            'value_space': self.value_space,
            'value_range': self.value_range,
            'value_note': self.value_note
        }


# 将完整的URL作为一个表名
def extract(source: list, target):
    cnt = 1
    schema_dict = {}
    for data in source:
        for line in load_jsonl(data):
            # 进行两轮抽取，第一次不解析，第二次解析
            question = entity_recognition(line['question'])
            for query in line['query']:
                schema_name = query['url']
                method = query['method']
                if schema_name not in schema_dict.keys():
                    now_schema = RawSchema()
                    now_schema.id = str(cnt)
                    now_schema.url = schema_name
                    now_schema.method = method.upper()
                    schema_dict[schema_name] = now_schema
                    cnt += 1
                else:
                    now_schema = schema_dict[schema_name]
                if 'params' not in query.keys():
                    continue
                for param in query['params']:
                    param_name = param['name'].strip()
                    if param_name == 'order':
                        continue
                    value_list = param['value'].replace('[', '').replace(']', '').replace('"', '').split(',')
                    flag = value_in_question(question, value_list)
                    if flag:
                        if param_name in now_schema.params.keys():
                            continue
                        now_param = RawParam()
                        now_param.param_name = param_name
                        now_param.param_note = ''
                        now_param.value_space = 'open'
                        if 'time' in param_name:
                            now_param.value_type = 'date'
                        elif 'amount' in param_name or 'number' in param_name:
                            now_param.value_type = 'number'
                        else:
                            now_param.value_type = 'string'
                        now_schema.params[param_name] = now_param
                    else:
                        if param_name in now_schema.params.keys():
                            now_param = now_schema.params[param_name]
                            for value in value_list:
                                if value not in now_param.value_range:
                                    now_param.value_range.append(value)
                                    now_param.value_note.append('')
                        else:
                            now_param = RawParam()
                            now_param.param_name = param_name
                            now_param.param_note = ''
                            now_param.value_type = 'string'
                            now_param.value_space = 'closed'
                            for value in value_list:
                                now_param.value_range.append(value)
                                now_param.value_note.append('')
                            now_schema.params[param_name] = now_param
    data = []
    for value in schema_dict.values():
        data.append(value.to_dict())
    dump_jsonl(data, target)


def value_in_question(question, value_list):
    flag = True
    for value in value_list:
        if value in question:
            continue
        else:
            flag = False
            break
    return flag



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import os
import numpy as np
from Text2JSON.featurizer.schema_featurizer import SchemaDataset
from Text2JSON.model.base_model import BaseModel


class SchemaEvaluator:
    def __init__(self, output_path, config, model: BaseModel, eval_schema_data):
        """
            评估网络效果
            output_path: 结果输出路径
            config: 参数配置
            eval_data: 需要验证的数据
            hydra_featurizer: 特征提取器
            model: 模型
            note: 注释
        """
        self.config = config
        self.model = model
        self.eval_history_file = os.path.join(output_path, "schema_eval.log")  # 预测结果存放位置
        f = open(self.eval_history_file, 'w')
        f.close()
        self.eval_schema_data = eval_schema_data

    # 改变这里的评测方式
    def eval_schema(self, eval_table_data: SchemaDataset):
        items = ['overall', 'first', 'first_contain']
        acc = {k: 0.0 for k in items}
        cnt = 0
        model_outputs = self.model.dataset_inference(eval_table_data)
        for input_feature, model_output in zip(eval_table_data.input_features, model_outputs):
            cur_acc = {'first': 1, 'first_contain': 1}

            schema_rel_label = input_feature.schema_rel

            if 1 not in input_feature.schema_mask:
                sql_seq = 1
            else:
                sql_seq = 2

            if sql_seq == 1:
                # parsing sql1 schema relevant
                schema_relevant_prob = np.exp(model_output['schema_first_sim'][:, 0])
                schema_id_prob_exp = sorted(enumerate(schema_relevant_prob), key=lambda x: x[1], reverse=True)
                schema_rel = []
                for table_index, table_pro in schema_id_prob_exp:
                    if table_pro >= float(self.config['schema_threshold']):
                        schema_rel.append(table_index)
                        if len(schema_rel) >= int(self.config['schema_remain']):
                            break
            else:
                # parsing sql2 schema relevant
                table_relevant_prob = np.exp(model_output['schema_sec_sim'][:, 0])
                table_id_prob_exp = sorted(enumerate(table_relevant_prob), key=lambda x: x[1], reverse=True)
                schema_rel = []
                for table_index, table_pro in table_id_prob_exp:
                    if table_pro >= float(self.config['schema_threshold']):
                        schema_rel.append(table_index)
                        if len(schema_rel) >= int(self.config['schema_remain']):
                            break

            # test schema relevant
            for index, value in enumerate(schema_rel_label):
                if value == 1 and (len(schema_rel) == 0 or index != schema_rel[0]):
                    cur_acc['first'] = 0
                if value == 1 and index not in schema_rel:
                    cur_acc['first_contain'] = 0
                    break

            # add all
            for k in cur_acc:
                acc[k] += cur_acc[k]

            all_correct = 0 if 0 in cur_acc.values() else 1
            cur_acc['overall'] = all_correct
            acc["overall"] += all_correct

            cnt += 1

        # set result
        result_str = []
        for item in items:
            result_str.append(item + ":{0:.1f}".format(acc[item] * 100.0 / cnt))
        result_str = ", ".join(result_str)
        return result_str, acc

    def eval(self, epochs):
        print('schema eval_file', self.config['dev_data_path'])
        result_str, acc = self.eval_schema(self.eval_schema_data)
        print("[ epochs " + str(epochs) + " ] " + result_str)
        with open(self.eval_history_file, "a+", encoding="utf8") as f:
            f.write("[{0}, epoch {1}] ".format(self.config['dev_data_path'], epochs) + result_str + "\n")
        return result_str, acc

    # 将最好结果写入
    def append(self, state, result_str):
        with open(self.eval_history_file, "a+", encoding="utf8") as f:
            f.write("[{0}, {1}] ".format(self.config['dev_data_path'], state) + result_str + "\n")







import sys
from os.path import dirname, abspath

path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
from Text2JSON.train import utils
from Text2JSON.predict.prediction import Predict
import datetime


def f_one(label_file, predict_file):
    stop_word_list = ['【', '】', '[', ']', '"', "'", '‘', '’', '“', '，', '.', '。', ' ', '”']
    label_data = utils.load_jsonl_by_key('qid', label_file)
    predit_data = utils.load_jsonl_by_key('qid', predict_file)
    sum_method_recall, sum_url_recall, sum_params_name_recall, \
    sum_params_op_recall, sum_params_value_recall, sum_rel_recall = 0., 0., 0., 0., 0., 0.

    sum_method_acc, sum_url_acc, sum_params_name_acc, \
    sum_params_op_acc, sum_params_value_acc, sum_rel_acc = 0., 0., 0., 0., 0., 0.

    for qid in label_data.keys():
        label = label_data[qid]
        predict = predit_data[qid]
        label_query = label['query']
        predict_query = predict['query']

        # predict parsing
        predict_params_name = {}
        predict_params_op = {}
        predict_params_value = {}
        for query in predict_query:
            url = query['url']
            name_to_option_dict = {}
            name_to_value_dict = {}
            predict_params_name[url] = []
            predict_params_op[url] = name_to_option_dict
            predict_params_value[url] = name_to_value_dict
            for params in query['params']:
                name = params['name']
                option = params['option']
                value = str(params['value'])
                for i in stop_word_list:
                    value = value.replace(i, '')
                value = value.split(',')
                predict_params_name[url].append(name)
                predict_params_op[url][name] = option
                predict_params_value[url][name] = value

        # label parsing
        label_params_name = {}
        label_params_op = {}
        label_params_value = {}
        for query in label_query:
            url = query['url']
            name_to_option_dict = {}
            name_to_value_dict = {}
            label_params_name[url] = []
            label_params_op[url] = name_to_option_dict
            label_params_value[url] = name_to_value_dict
            if 'params' not in query.keys():
                continue
            for params in query['params']:
                name = params['name']
                option = params['option']
                value = str(params['value'])
                for i in stop_word_list:
                    value = value.replace(i, '')
                value = value.split(',')
                label_params_name[url].append(name)
                label_params_op[url][name] = option
                label_params_value[url][name] = value

        rel_recall = 0.
        rel_acc = 0.
        # eval structure message
        if len(label_query) == len(predict_query):
            rel_acc = 1
            rel_recall = 1
        if len(label_query) > len(predict_query):
            rel_recall = (float(len(predict_query)) + 0.0001) / (len(label_query) + 0.0001)
            if len(predict_query) == 0:
                rel_acc = 0
            else:
                rel_acc = 1
        if len(predict_query) > len(label_query):
            rel_acc = (float(len(label_query)) + 0.0001) / (len(predict_query) + 0.0001)
            rel_recall = 1
        # eval url and method
        url_num = 0
        label_url = label_params_name.keys()
        predict_url = predict_params_name.keys()
        for url in label_url:
            if url in predict_url:
                url_num += 1

        url_acc = (url_num + 0.0001) / (float(len(predict_url)) + 0.0001)
        url_recall = (url_num + 0.0001) / (float(len(label_url)) + 0.0001)
        method_acc = url_acc
        method_recall = url_recall

        # eval params_name recall
        params_name_recall = 0
        for url, label_name_list in label_params_name.items():
            sub_name_num = 0
            if url not in predict_params_name.keys():
                continue
            predict_name_list = predict_params_name[url]

            for param in label_name_list:
                if param in predict_name_list:
                    sub_name_num += 1
            params_name_recall += (sub_name_num + 0.0001) / (float(len(label_name_list)) + 0.0001)
        params_name_recall = (params_name_recall + 0.0001) / (float(len(label_params_name.keys())) + 0.0001)

        # params acc
        params_name_acc = 0
        for url, predict_name_list in predict_params_name.items():
            sub_name_num = 0
            if url not in label_params_name.keys():
                continue
            label_name_list = label_params_name[url]
            for param in predict_name_list:
                if param in label_name_list:
                    sub_name_num += 1
            params_name_acc += (sub_name_num + 0.0001) / (float(len(predict_name_list)) + 0.0001)
        params_name_acc = (params_name_acc + 0.0001) / (float(len(predict_params_name.keys())) + 0.0001)

        # option recall
        params_op_recall = 0
        for url, label_option_dict in label_params_op.items():
            sub_op_num = 0
            if url not in predict_params_op.keys():
                continue
            predict_option_dict = predict_params_op[url]
            for key, op in label_option_dict.items():
                if key in predict_option_dict.keys():
                    if op == predict_option_dict[key]:
                        sub_op_num += 1
            params_op_recall += (sub_op_num + 0.0001) / (float(len(label_option_dict.keys())) + 0.0001)
        params_op_recall = (params_op_recall + 0.0001) / (float(len(label_params_op.keys())) + 0.0001)

        # option acc
        params_op_acc = 0
        for url, predict_option_dict in predict_params_op.items():
            sub_op_num = 0
            if url not in label_params_name.keys():
                continue
            label_option_dict = label_params_op[url]
            for key, op in predict_option_dict.items():
                if key in label_option_dict.keys():
                    if op == label_option_dict[key]:
                        sub_op_num += 1
            params_op_acc += (sub_op_num + 0.0001) / (float(len(predict_option_dict.keys())) + 0.0001)
        params_op_acc = (params_op_acc + 0.0001) / (float(len(predict_params_op.keys())) + 0.0001)

        # value recall
        params_value_recall = 0
        for url, label_value_dict in label_params_value.items():
            if url not in predict_params_value.keys():
                continue
            predict_value_dict = predict_params_value[url]
            sec_layer_recall = 0
            for key, value_list in label_value_dict.items():
                if key not in predict_value_dict:
                    continue
                value_num = 0
                for value in value_list:
                    if value in predict_value_dict[key]:
                        value_num += 1
                sec_layer_recall += (value_num + 0.0001) / (float(len(value_list) + 0.0001))
            params_value_recall += (sec_layer_recall + 0.0001) / (float(len(label_value_dict.keys()) + 0.0001))
        params_value_recall = (params_value_recall + 0.0001) / (float(len(label_params_value.keys()) + 0.0001))

        # value acc
        params_value_acc = 0
        for url, predict_value_dict in predict_params_value.items():
            if url not in label_params_value.keys():
                continue
            label_value_dict = label_params_value[url]
            sec_layer_acc = 0
            for key, value_list in predict_value_dict.items():
                if key not in label_value_dict:
                    continue
                value_num = 0
                for value in value_list:
                    if value in label_value_dict[key]:
                        value_num += 1
                sec_layer_acc += (value_num + 0.0001) / (float(len(value_list) + 0.0001))
            params_value_acc += (sec_layer_acc + 0.0001) / (float(len(predict_value_dict.keys()) + 0.0001))
        params_value_acc = (params_value_acc + 0.0001) / (float(len(predict_params_value.keys()) + 0.0001))

        sum_method_recall += method_recall
        sum_url_recall += url_recall
        sum_params_name_recall += params_name_recall
        sum_params_op_recall += params_op_recall
        sum_params_value_recall += params_value_recall
        sum_rel_recall += rel_recall

        sum_method_acc += method_acc
        sum_url_acc += url_acc
        sum_params_name_acc += params_name_acc
        sum_params_op_acc += params_op_acc
        sum_params_value_acc += params_value_acc
        sum_rel_acc += rel_acc

    cnt = len(label_data.keys())
    sum_method_recall = sum_method_recall / cnt
    sum_url_recall = sum_url_recall / cnt
    sum_params_name_recall = sum_params_name_recall / cnt
    sum_params_op_recall = sum_params_op_recall / cnt
    sum_params_value_recall = sum_params_value_recall / cnt
    sum_rel_recall = sum_rel_recall / cnt

    sum_method_acc = sum_method_acc / cnt
    sum_url_acc = sum_url_acc / cnt
    sum_params_name_acc = sum_params_name_acc / cnt
    sum_params_op_acc = sum_params_op_acc / cnt
    sum_params_value_acc = sum_params_value_acc / cnt
    sum_rel_acc = sum_rel_acc / cnt

    sum_recall = (sum_method_recall + sum_url_recall + sum_params_name_recall
                  + sum_params_op_recall + sum_params_value_recall + sum_rel_recall) / 6.

    sum_acc = (sum_method_acc + sum_url_acc + sum_params_name_acc + sum_params_op_acc
               + sum_params_value_acc + sum_rel_acc) / 6.

    f_one = (2 * sum_recall * sum_acc) / (sum_recall + sum_acc)
    result_str = 'F1:{0:.4f}, Recall:{1:.4f}, ACC:{2:.4f}'.format(f_one, sum_recall, sum_acc) + '\n' + \
                 '-------------------------------------------------------------------------------------' + '\n' + \
                 'Recall' + '\n' + \
                 'Method:{0:.4f}, URL:{1:.4f}, Name:{2:.4f}, Option:{3:.4f}, Value:{4:.4f}, Rel:{5:.4f}'. \
                     format(sum_url_recall, sum_url_recall, sum_params_name_recall, sum_params_op_recall,
                            sum_params_value_recall, sum_rel_recall) + '\n' + \
                 '-------------------------------------------------------------------------------------' + '\n' + \
                 'Accuracy' + '\n' + \
                 'Method:{0:.4f}, URL:{1:.4f}, Name:{2:.4f}, Option:{3:.4f}, Value:{4:.4f}, Rel:{5:.4f}'. \
                     format(sum_url_acc, sum_url_acc, sum_params_name_acc, sum_params_op_acc, sum_params_value_acc,
                            sum_rel_acc) + '\n'

    print('F1:{0:.4f}, Recall:{1:.4f}, ACC:{2:.4f}'.format(f_one, sum_recall, sum_acc))
    print('----------------------------------------------------------------------------------')
    print('Recall')
    print('Method:{0:.4f}, URL:{1:.4f}, Name:{2:.4f}, Option:{3:.4f}, Value:{4:.4f}, Rel:{5:.4f}'.format(sum_url_recall,
                                                                                                         sum_url_recall,
                                                                                                         sum_params_name_recall,
                                                                                                         sum_params_op_recall,
                                                                                                         sum_params_value_recall,
                                                                                                         sum_rel_recall))
    print('----------------------------------------------------------------------------------')
    print('Accuracy')
    print('Method:{0:.4f}, URL:{1:.4f}, Name:{2:.4f}, Option:{3:.4f}, Value:{4:.4f}, Rel:{5:.4f}'.
          format(sum_url_acc, sum_url_acc, sum_params_name_acc, sum_params_op_acc, sum_params_value_acc, sum_rel_acc))

    return result_str


def bleu_one(label_file, predict_file):
    stop_word_list = ['【', '】', '[', ']', '"', "'", '‘', '’', '“', '，', '.', '。', ' ', '”']
    label_data = utils.load_jsonl_by_key('qid', label_file)
    predit_data = utils.load_jsonl_by_key('qid', predict_file)
    sum_bleu = 0
    for qid in label_data.keys():
        label = label_data[qid]
        predict = predit_data[qid]
        label_query = label['query']
        predict_query = predict['query']

        # predict parsing
        predict_list = []
        for query in predict_query:
            url = query['url']
            method = query['method']
            predict_list.append(url)
            predict_list.append(method)
            for params in query['params']:
                name = params['name']
                option = params['option']
                value = str(params['value'])
                for i in stop_word_list:
                    value = value.replace(i, '')
                value = value.split(',')
                predict_list.append(name)
                predict_list.append(option)
                predict_list.extend(value)

        # label parsing
        label_list = []
        for query in label_query:
            url = query['url']
            method = query['method']
            label_list.append(url)
            label_list.append(method)
            if 'params' not in query.keys():
                continue
            for params in query['params']:
                name = params['name']
                option = params['option']
                value = str(params['value'])
                for i in stop_word_list:
                    value = value.replace(i, '')
                value = value.split(',')
                label_list.append(name)
                label_list.append(option)
                label_list.extend(value)

        bleu = 0
        for token in predict_list:
            if token in label_list:
                bleu += 1

        sum_bleu += float(bleu) / len(predict_list)

    sum_bleu = sum_bleu / len(predit_data.keys())
    result_str = 'BLEU-1:{0:.4f}'.format(sum_bleu) + \
                 '\n----------------------------------------------------------------------'
    print('BLEU-1:{0:.4f}'.format(sum_bleu))
    return result_str


def eval_bleu1_f1(model: Predict, label_path, output_path):
    """
        param: model 客户端模型
        label_path:
    """
    predict_list = []
    for data in utils.load_jsonl(label_path):
        qid = data['qid']
        question = data['question']
        predict_list.append(model.predict(question, qid=qid))

    utils.dump_jsonl(predict_list, output_path)
    a_result = bleu_one(label_path, output_path)
    b_result = f_one(label_path, output_path)
    return a_result, b_result



import json


class SQLExample(object):
    def __init__(self):
        """
            这是拆分过后SQL
        """
        self.qid = None
        self.table_id = None
        self.table_id_relative = None
        self.question = None
        self.tokens = None
        self.char_to_word = None
        self.word_to_char_start = None

        self.seq = None
        self.agg = None
        self.sel = None
        self.conds = None
        self.rel = None
        self.value_start_end = None
        self.relationship = None
        self.valid = None

        # 预测时记载数据时用的
        self.first_schema_id = []
        self.sec_schema_id = []

    # 静态方法
    @staticmethod
    def load_from_json(s):
        data = json.loads(s)
        example_list = []
        qid = data['qid']
        question = data['question']
        tokens = data['tokens']
        char_to_word = data['char_to_word']
        word_to_char_start = data['word_to_char_start']
        valid = data['valid']
        for index in range(0, len(data['sql']['clause'])):
            table_id = data['table_id'][index]
            table_id_relative = data['table_id'][:index]+data['table_id'][index+1:]
            seq = data['sql']['clause'][index]['seq']
            sel = data['sql']['clause'][index]['sel']
            agg = data['sql']['clause'][index]['agg']
            conds = data['sql']['clause'][index]['conds']
            value_start_end = data['sql']['clause'][index]['value_start_end']
            relationship = data['sql']['relationship']

            rel = []
            rel.extend(sel)
            for cond in conds:
                rel.append(cond[0])

            example = SQLExample()
            example.qid = qid
            example.table_id = table_id
            example.table_id_relative = table_id_relative
            example.question = question
            example.tokens = tokens
            example.char_to_word = char_to_word
            example.word_to_char_start = word_to_char_start
            example.seq = seq
            example.sel = sel
            example.agg = agg
            example.conds = conds
            example.rel = rel
            example.value_start_end = value_start_end
            example.relationship = relationship
            example.valid = valid
            example_list.append(example)
        return example_list

if __name__ == "__main__":
    ''


import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import datetime
import re
from datetime import timedelta
from Text2JSON.entity_named_recog.hour_entity_recognition import hour_absolute_comfirm
from Text2JSON.entity_named_recog.entity_utils import *


def minute_recon(line, placeholders_list: dict):
    line = minute_relative_recognition(line, placeholders_list)
    line = minute_absolute_recognition(line, placeholders_list)
    return line


def minute_absolute_recognition(line, placeholders_list: dict):
    now = datetime.datetime.now()
    time_map = {'零': '0', '一': '1', '二': '2', '三': '3', '四': '4',
                '五': '5', '六': '6', '七': '7', '八': '8', '九': '9', '十': '10', '两': '2', '半':'30'}
    # 数字 分钟
    date_all = re.findall(r"([:点]\d{1,2}[分]?[钟]?)", line)
    for data in date_all:
        data = data[1:]
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        minute = re.sub(r'([分]?[钟]?)', '', data)

        # 确定年月日
        handle_line, year, month, day, hour = hour_absolute_comfirm(first_line, placeholders_list)

        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day
        hour = now.hour if hour is None else hour

        # 确定分钟
        minute = int(minute)
        try:
            if minute > 60:
                raise ValueError
            if hour > 24:
                raise ValueError

            target_time = datetime.datetime(year, month, day, 0, 0)
            target_time = target_time + timedelta(minutes=minute)
            target_time = target_time + timedelta(hours=hour)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 数字 分钟
    date_all = re.findall(r"([:点][二三四五六七八九]?[十]?[零一二三四五六七八九十半][分]?[钟]?)", line)
    for data in date_all:
        data = data[1:]
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        minute = re.sub(r'([分]?[钟]?)', '', data)

        # 确定年月日
        handle_line, year, month, day, hour = hour_absolute_comfirm(first_line, placeholders_list)

        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day
        hour = now.hour if hour is None else hour

        # 解析分钟
        if len(minute) > 1:
            if '十' not in minute:
                # 错误处理
                minute = 61
            else:
                pre_minute = minute[0:minute.index('十')]
                post_minute = minute[minute.index('十') + 1:]
                # 10
                if len(pre_minute) == 0 and len(post_minute) == 0:
                    minute = 10
                # 11 ~ 19
                elif len(pre_minute) == 0 and len(post_minute) != 0:
                    minute = 10 + int(time_map[post_minute])
                # 20, 30
                elif len(pre_minute) != 0 and len(post_minute) == 0:
                    minute = int(time_map[pre_minute]) * 10
                else:
                    # 21 ~ 24
                    minute = int(time_map[pre_minute]) * 10 + int(time_map[post_minute])
        else:
            minute = int(time_map[minute])

        # 确定分钟
        try:
            if minute > 60:
                raise ValueError
            if hour > 24:
                raise ValueError

            target_time = datetime.datetime(year, month, day, 0, 0)
            target_time = target_time + timedelta(minutes=minute)
            target_time = target_time + timedelta(hours=hour)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line
    return line


def minute_relative_recognition(line, placeholders_list: dict):
    time_map = {'一': '1', '二': '2', '三': '3', '四': '4',
                '五': '5', '六': '6', '七': '7', '八': '8', '九': '9', '两': '2', '十': '10'}
    # 半单独拿出来，放到分钟那里
    now = datetime.datetime.now()
    # 中文前 1-99
    date_all = re.findall(r"([二三四五六七八九]?[十]?[一两二三四五六七八九十]分钟前)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        # 计算时差，获取目标时间
        minute = re.sub(r"分钟前", '', data)
        try:
            # 解析分钟
            if len(minute) > 1:
                if '十' not in minute:
                    # 错误处理
                    raise ValueError
                else:
                    pre_minute = minute[0:minute.index('十')]
                    post_minute = minute[minute.index('十') + 1:]
                    # 10
                    if len(pre_minute) == 0 and len(post_minute) == 0:
                        minute = 10
                    # 11 ~ 19
                    elif len(pre_minute) == 0 and len(post_minute) != 0:
                        minute = 10 + int(time_map[post_minute])
                    # 20, 30
                    elif len(pre_minute) != 0 and len(post_minute) == 0:
                        minute = int(time_map[pre_minute]) * 10
                    else:
                        # 21 ~ 24
                        minute = int(time_map[pre_minute]) * 10 + int(time_map[post_minute])
            else:
                minute = int(time_map[minute])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        target_time = now - timedelta(minutes=minute)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line

    # 数字前
    date_all = re.findall(r"(\d{1,2}分钟前)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        # 计算时差，获取目标时间
        data = re.sub(r"分钟前", '', data)
        minute_delta = int(data)
        target_time = now - timedelta(minutes=minute_delta)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line

    # 中文后 10-99
    date_all = re.findall(r"([二三四五六七八九]?[十]?[一两二三四五六七八九十]分钟后)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        # 计算时差，获取目标时间
        minute = re.sub(r"分钟后", '', data)

        try:
            # 解析分钟
            if len(minute) > 1:
                if '十' not in minute:
                    raise ValueError
                else:
                    pre_minute = minute[0:minute.index('十')]
                    post_minute = minute[minute.index('十') + 1:]
                    # 10
                    if len(pre_minute) == 0 and len(post_minute) == 0:
                        minute = 10
                    # 11 ~ 19
                    elif len(pre_minute) == 0 and len(post_minute) != 0:
                        minute = 10 + int(time_map[post_minute])
                    # 20, 30
                    elif len(pre_minute) != 0 and len(post_minute) == 0:
                        minute = int(time_map[pre_minute]) * 10
                    else:
                        # 21 ~ 24
                        minute = int(time_map[pre_minute]) * 10 + int(time_map[post_minute])
            else:
                minute = int(time_map[minute])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue

        target_time = now + timedelta(minutes=minute)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line

    # 数字后
    date_all = re.findall(r"(\d{1,2}分钟后)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        # 计算时差，获取目标时间
        data = re.sub(r"分钟后", '', data)
        minute_delta = int(data)
        target_time = now + timedelta(minutes=minute_delta)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line
    return line



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
from Text2JSON.train import utils
import re


class schema_parsed(object):
    def __init__(self):
        self.extend_schema = {}
        # 局部列序号到列全局序号的相互映射
        self.global_to_local = {}
        self.local_to_global = {}
        # 局部表序号到表全局序号的相互映射
        self.table_global_to_local = {}
        self.table_local_to_global = {}
        self.url_method = {}
        self.column_length = 0

    def load_schema(self, config, tokenizer):
        self.extend_schema = utils.load_jsonl_by_key('id', config['schema_path'])
        self.extend_schema_feature(tokenizer)
        self.extend_schema_feature_without_note(tokenizer)
        self.extend_schema_table_feature(tokenizer)
        self.extend_schema_table_feature_without_note(tokenizer)

        self.global_to_local, self.local_to_global = self.parse_column_map(self.extend_schema)
        self.table_global_to_local, self.table_local_to_global = self.parse_table_map(self.extend_schema)

        self.url_method = self.url_mathod_map(self.extend_schema)
        self.column_length = len(self.global_to_local.keys())

    def save_schema(self, config):
        utils.dump_model(self, config['featured_schema_path'])
        print('{0} file have been featured and the featured file saved in {1}'.
              format(config["schema_path"], config['featured_schema_path']))

    def parse_table_map(self, extend_schema):
        """
            local_global_map : table_id + col_id -> global_col_id 将局部col_id 映射到全局col_id
            global_local_map : global_col_id -> table_id + col_id 将全局col_id 映射到局部col_id
        """
        table_global_local_map = {}
        table_local_global_map = {}
        cur_id = 0
        for table_id, value in extend_schema.items():
            table_local_global_map[table_id] = cur_id
            table_global_local_map[cur_id] = table_id
            cur_id += 1
        return table_global_local_map, table_local_global_map

    def parse_column_map(self, extend_schema):
        """
            local_global_map : table_id + col_id -> global_col_id 将局部col_id 映射到全局col_id
            global_local_map : global_col_id -> table_id + col_id 将全局col_id 映射到局部col_id
        """
        global_local_map = {}
        local_global_map = {}
        cur_id = 0
        for table_id, value in extend_schema.items():
            for col_id in range(0, len(value['header'])):
                local_id = str(table_id)+"/"+str(col_id)
                local_global_map[local_id] = cur_id
                global_local_map[cur_id] = str(table_id)+"/"+str(col_id)
                cur_id += 1
        return global_local_map, local_global_map

    def extend_schema_feature_without_note(self, tokenizer):
        # extend feature
        # table_name table_note, column_name, column_type, column_note, column_space
        for key, value in self.extend_schema.items():
            feature_list = []
            table_name = value['url']
            # table_note = value['table_note']
            table_name_list = re.split('[/_]', table_name)
            # table_note_list = re.split('[,]', table_note)
            for index in range(0, len(value['header'])):
                sub_token = []
                column_name = value['header'][index]
                column_type = value['types'][index]
                # column_note = value['notes'][index]
                column_space = value['space'][index]

                column_name_list = re.split('[&_/]', column_name)
                column_type_list = re.split('[,]', column_type)
                # column_note_list = re.split('[,]', column_note)
                column_space_list = re.split('[,]', column_space)

                sub_token.extend(table_name_list)
                # sub_token.extend(table_note_list)
                sub_token.extend(column_name_list)
                sub_token.extend(column_type_list)
                # sub_token.extend(column_note_list)
                sub_token.extend(column_space_list)
                sequence = ' '.join(sub_token).strip()
                feature_list.append(tokenizer.tokenize(sequence))
            value['feature_column_without_note'] = feature_list

    def extend_schema_feature(self, tokenizer):
        # extend feature
        # table_name table_note, column_name, column_type, column_note, column_space
        for key, value in self.extend_schema.items():
            feature_list = []
            table_name = value['url']
            table_note = value['url_note']
            table_name_list = re.split('[/_]', table_name)
            table_note_list = re.split('[,]', table_note)
            for index in range(0, len(value['header'])):
                sub_token = []
                column_name = value['header'][index]
                column_type = value['types'][index]
                column_note = value['notes'][index]
                column_space = value['space'][index]

                column_name_list = re.split('[&_/]', column_name)
                column_type_list = re.split('[,]', column_type)
                column_note_list = re.split('[,]', column_note)
                column_space_list = re.split('[,]', column_space)

                sub_token.extend(table_name_list)
                sub_token.extend(table_note_list)
                sub_token.extend(column_name_list)
                sub_token.extend(column_type_list)
                sub_token.extend(column_note_list)
                sub_token.extend(column_space_list)
                sequence = ' '.join(sub_token).strip()
                feature_list.append(tokenizer.tokenize(sequence))
            value['feature_column'] = feature_list

    def extend_schema_table_feature(self, tokenizer):
        meanless_list = ['v2']
        # extend feature
        # table_name table_note, column_name, column_type, column_note, column_space
        for key, value in self.extend_schema.items():
            feature_list = []
            table_name = value['url']
            table_note = value['url_note']
            table_name_list = re.split('[/_]', table_name)
            for stop_word in meanless_list:
                if stop_word in table_name_list:
                    table_name_list.remove(stop_word)
            table_note_list = re.split('[,]', table_note)
            sub_token = []
            sub_token.extend(table_name_list)
            sub_token.extend(table_note_list)
            sequence = ' '.join(sub_token).strip()
            feature_list.append(tokenizer.tokenize(sequence))
            value['feature_schema'] = feature_list

    def extend_schema_table_feature_without_note(self, tokenizer):
        meanless_list = ['v2']
        # extend feature
        # table_name table_note, column_name, column_type, column_note, column_space
        for key, value in self.extend_schema.items():
            feature_list = []
            table_name = value['url']
            table_name_list = re.split('[/_]', table_name)
            for stop_word in meanless_list:
                if stop_word in table_name_list:
                    table_name_list.remove(stop_word)
            sub_token = []
            sub_token.extend(table_name_list)
            sequence = ' '.join(sub_token).strip()
            feature_list.append(tokenizer.tokenize(sequence))
            value['feature_schema_without_note'] = feature_list

    def url_mathod_map(self, extend_schema):
        url_method = {}
        for key in extend_schema:
            schema = extend_schema[key]
            url_method[schema['url']] = schema['method']
        return url_method

    def to_dict(self):
        return {
            'local_to_global': self.local_to_global,
            'global_to_global': self.global_to_local
        }



import os
import json
import transformers
import pickle


def load_jsonl(jsonl):
    for line in open(jsonl, encoding="utf8"):
        sample = json.loads(line.rstrip())
        yield sample


def dump_jsonl(data, path):
    with open(path, 'w', encoding='utf-8') as file_obj:
        for line in data:
            file_obj.writelines(json.dumps(line, ensure_ascii=False) + '\n')


def load_jsonl_by_key(key, filepath):
    table_dict = {}
    with open(filepath, 'r', encoding='utf8')as fp:
        for line in fp:
            line_json = json.loads(line)
            table_dict[line_json[key]] = line_json
        return table_dict


def load_json(filepath):
    with open(filepath, 'r', encoding='utf8')as fp:
        return json.load(fp)


def dump_json(data, filepath):
    with open(filepath, 'w', encoding='utf-8') as file_obj:
        json.dump(data, file_obj, ensure_ascii=False)


def dump_model(model, path):
    with open(path, 'wb') as f:
        pickle.dump(model, f)


def load_model(path):
    with open(path, 'rb') as f:
        return pickle.load(f)


def dump_probablity(data, path):
    with open(path, 'w', encoding='utf-8') as file_obj:
        for qid, line in data.items():
            file_obj.writelines(str(qid) + ' ' + str(line) + '\n')


def read_conf(conf_path):
    config = {}
    for line in open(conf_path, encoding="utf8"):
        line = line.strip()
        if line.strip() == "" or line[0] == "#":
            continue
        try:
            sharp_index = line.index('#')
        except:
            sharp_index = len(line)
        fields = line[:sharp_index].split(" ")
        config[fields[0]] = fields[1]
    return config


def create_base_model(config):
    try:
        return transformers.BertModel.from_pretrained(config['bert_path'])
    except Exception:
        raise Exception("base_path {0} not supported".format(config['bert_path']))


def create_tokenizer(config):
    try:
        return transformers.BertTokenizer.from_pretrained(config['bert_path'])
    except Exception:
        raise Exception("base_path {0} not supported".format(config['bert_path']))



import sys
from os.path import dirname, abspath

path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import numpy as np
import torch.utils.data as torch_data
from Text2JSON.featurizer.schema_parser import schema_parsed
from Text2JSON.featurizer.Input_SQL import SQLExample
from collections import defaultdict

stats = defaultdict(int)


class ColumnInputFeature(object):
    def __init__(self,
                 qid,
                 question,
                 table_id,
                 tokens,
                 word_to_char_start,
                 word_to_subword,
                 subword_to_word,
                 input_ids,
                 input_mask,
                 segment_ids):
        """
            qid: question id
            question: question token 单个
            table_id: table id 单个
            tokens: [[cls]type,column[SEP]question[SEP],...] 一组 column和question的联合表示，获取编码后的序列
            word_to_char_start: [start1, start2, start3, start4] 一组 记录问题token字符的开始位置
            word_to_subword: [[(start_index, end_index),...,],...,[]] 一组 词 到 子词 的映射关系,形式
            subword_to_word: [[0,1,1,2,2,2],...] 一组 子词 到 词 的映射关系
            input_ids: [[id1, id2, id3,..., pad, pad],...] 一组 将输入tokens, 通过词表转换成id
            input_mask: [[1,1,1,1],...] 一组 标识column+question位置编码,也就是整个输入所占的编码
            segment_ids: [[0,0,1,1],...] 一组 标识column和question位置
            columns: [name1, name2, name3, name4, name5] question有可能所涉及到的列
            seq: [0, 0, 0, 0, 0] 位置信息0代表第一条SQL语句，1代表第二条SQL语句
            sel_num: [1, 1, 1, 1, 1] select_clause 对应列的数目
            sel: [0,0,1,0,0] select_clause 所选择的列
            agg: [0, 0, 1, 0, 0]有可能所涉及到的列对应操作符
            where_num: [1, 1, 1, 1, 1] 每一个<>column<>question<> pair 对应的列数目, [len(conds)] * columns
            where: [0, 0, 1, 0, 0] 记录每一个列是否出现在where_clause中
            op: [0, 0, 2, 0, 0] 记录在where_clause中，每一个列对应的操作符
            value_num: [1, 2, 3, 0, 0] where子句中列对应值的数目
            value_start: [[0,...], [0,...], [start_pos1, start_pos2,...], ...] 记录在where_clause中，每一个列对应值在tokens的开始位置, 如果没有，记为-1
            value_end: [[0], [0], [end_pos1, end_pos2] , [0], [0]] 记录在where_clause中，每一个列对应值在tokens的结束位置, 如果没有，记为-1
            relationship: [0, 0, 0, 0, 0] 结构信息 0:代表单条SQL 1:代表两条条SQL
        """
        self.qid = qid
        self.question = question
        self.table_id = table_id
        self.tokens = tokens

        # 这三个都可能没有用
        self.word_to_char_start = word_to_char_start
        self.word_to_subword = word_to_subword
        self.subword_to_word = subword_to_word

        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids

        self.seq = None
        self.sel_num = None
        self.sel = None
        self.agg = None
        self.where_num = None
        self.where = None
        self.op = None
        self.value_num = None
        self.value_start = None
        self.value_end = None
        self.relationship = None

        self.tokens_common = None

    def to_dict(self):
        return {
            'qid': self.qid,
            'question': self.question,
            'table_id': self.table_id,
            'tokens': self.tokens,
            'word_to_char_start': self.word_to_char_start,
            'word_to_subword': self.word_to_subword,
            'subword_to_word': self.subword_to_word,
            'input_ids': self.input_ids,
            'input_mask': self.input_mask,
            'segment_ids': self.segment_ids,

            'seq': self.seq,
            'sel_num': self.sel_num,
            'sel': self.sel,
            'agg': self.agg,
            'where_num': self.where_num,
            'where': self.where,
            'op': self.op,
            'value_num': self.value_num,
            'value_start': self.value_start,
            'value_end': self.value_end,
            'relationship': self.relationship
        }


class HydraColumnFeaturizer(object):
    """
        特征提取器，将一条训练数据集，转变成为InputFeature
    """

    def __init__(self, config, tokenizer):
        self.config = config
        # 根据不同模型，设置tokenizer
        self.tokenizer = tokenizer

    def get_predict_input_column_feature(self, example: SQLExample, schema: schema_parsed, config, note=True):
        """
            example 是一条完整的train数据
            example:
            {
                table_id,
                question,
                tokens,
                word_to_char_start,
                column_meta
                ...
                只用到这些
            }
            schema:
            {
                id,
                name,
                table_note,
                header,
                types,
                space,
                notes,
                feature  # 只用到feature
            }
            填充InputFeature其中的九个值
            InputFeature:
            {
                question,
                table_id,
                word_to_char_start,
                tokens,
                word_to_subword,
                subword_to_word,
                input_ids,
                input_mask,
                segment_ids
            }
        """
        max_query_length = int(config["max_query_length"])
        max_column_length = int(config["max_column_length"])
        max_total_length = max_column_length + max_query_length

        # 一条feature包含一条train,包含多个<>column<>question<>编码
        input_feature = ColumnInputFeature(
            example.qid,
            example.question,
            example.table_id,
            [],
            example.word_to_char_start,
            [],
            [],
            [],
            [],
            []
        )

        # 用于记录问题
        input_feature.tokens_common = example.tokens

        # 记录映射关系
        tokens = []
        word_to_subword_base = []  # 词 到 子词 的映射关系,形式(start_index, end_index)
        subword_to_word_base = []  # 子词 到 词 的映射关系,形式(0,1,1,2,2,2)
        for i, query_token in enumerate(example.tokens):
            sub_tokens = self.tokenizer.tokenize(query_token)
            cur_pos = len(tokens)
            if len(sub_tokens) > 0:
                word_to_subword_base += [(cur_pos, cur_pos + len(sub_tokens))]
                tokens.extend(sub_tokens)
                subword_to_word_base.extend([i] * len(sub_tokens))

        # 是否需要对输入进行拓展
        if note:
            feature_type = 'feature_column'
        else:
            feature_type = 'feature_column_without_note'

        # 子加载与表对应的数据
        table = schema.extend_schema[example.table_id]
        for table_column_feature in table[feature_type]:
            # 编码，输入为type, table, column, table_note, column_note, space, token
            # 输出[cls]type, table, column, table_note, column_note, space[SEP]question[SEP]
            assert len(table_column_feature) <= max_column_length, "列特征长度过小，请将配置文件中的max_column_length增大"
            assert len(''.join(tokens)) <= max_query_length, "输入问句长度大于最大字符长度，请将配置文件中的max_query_length" \
                                                             "增大或者缩短输入问句长度"
            tokenize_result = self.tokenizer.encode_plus(
                table_column_feature,
                tokens,
                max_length=max_total_length,
                truncation_strategy="longest_first",
                pad_to_max_length=True,
            )

            input_ids = tokenize_result["input_ids"]  # 将输入tokens, 通过词表转换成id
            segment_ids = tokenize_result["token_type_ids"]  # 标识column和question位置[0,0,1,1]
            input_mask = tokenize_result["attention_mask"]  # 标识column+question位置编码,也就是整个输入所占的编码
            new_tokens = self.tokenizer.convert_ids_to_tokens(input_ids)  # 获取编码后的序列[cls]column[SEP]question[SEP]

            # 0 1 划分方式
            # [cls] column [sep] question [sep]
            #   0     0      0      1       1
            # 获取column+[sep]的长度
            column_token_length = 0
            for i, sid in enumerate(segment_ids):
                if sid == 1:
                    column_token_length = i
                    break

            # 计算编码后的subword_to_word，需要加上column_token_length偏移
            subword_to_word = [0] * column_token_length + subword_to_word_base
            # 计算编码后的word_to_subword，需要加上column_token_length偏移
            word_to_subword = [(pos[0] + column_token_length, pos[1] + column_token_length) for pos in
                               word_to_subword_base]

            assert len(input_ids) == max_total_length
            assert len(input_mask) == max_total_length
            assert len(segment_ids) == max_total_length

            input_feature.tokens.append(new_tokens)
            input_feature.word_to_subword.append(word_to_subword)
            input_feature.subword_to_word.append(subword_to_word)
            input_feature.input_ids.append(input_ids)
            input_feature.input_mask.append(input_mask)
            input_feature.segment_ids.append(segment_ids)

        return input_feature

    def get_input_column_feature(self, example: SQLExample, schema: schema_parsed, config, positive=True, negative=True,
                                 note=False):
        """
            example 是一条完整的train数据
            example:
            {
                table_id,
                question,
                tokens,
                word_to_char_start,
                column_meta
                ...
                只用到这些
            }
            schema:
            {
                id,
                name,
                table_note,
                header,
                types,
                space,
                notes,
                feature  # 只用到feature
            }
            填充InputFeature其中的九个值
            InputFeature:
            {
                question,
                table_id,
                word_to_char_start,
                tokens,
                word_to_subword,
                subword_to_word,
                input_ids,
                input_mask,
                segment_ids
            }
        """
        max_query_length = int(config["max_query_length"])
        max_column_length = int(config["max_column_length"])
        max_total_length = max_column_length + max_query_length

        # 一条feature包含一条train,包含多个<>column<>question<>编码
        input_feature = ColumnInputFeature(
            example.qid,
            example.question,
            example.table_id,
            [],
            example.word_to_char_start,
            [],
            [],
            [],
            [],
            []
        )

        # 记录映射关系
        tokens = []
        word_to_subword_base = []  # 词 到 子词 的映射关系,形式(start_index, end_index)
        subword_to_word_base = []  # 子词 到 词 的映射关系,形式(0,1,1,2,2,2)
        for i, query_token in enumerate(example.tokens):
            sub_tokens = self.tokenizer.tokenize(query_token)
            cur_pos = len(tokens)
            if len(sub_tokens) > 0:
                word_to_subword_base += [(cur_pos, cur_pos + len(sub_tokens))]
                tokens.extend(sub_tokens)
                subword_to_word_base.extend([i] * len(sub_tokens))

        # positive 只加载正例， negative 只加载负例， 两个都为true，一起加载
        # 遍历表中所有columns，为每一个column生成<>column<>question<>编码
        # 这里只要是表相关的都作为正例
        rel_list = example.rel
        rel_table = example.table_id
        # 是否需要对输入进行拓展
        if note:
            feature_type = 'feature_column'
        else:
            feature_type = 'feature_column_without_note'

        for key, value in schema.extend_schema.items():
            # 只要正例
            if positive and not negative:
                if key != rel_table:
                    continue
            # 只要负例
            elif not positive and negative:
                if key == rel_table:
                    continue
            # 正例和负例都要
            elif positive and negative:
                ''
            table = schema.extend_schema[key]
            for table_column_feature in table[feature_type]:
                # 编码，输入为type, table, column, table_note, column_note, space, token
                # 输出[cls]type, table, column, table_note, column_note, space[SEP]question[SEP]
                assert len(
                    table_column_feature) <= max_column_length, "{0} URL中参数特征长度过大，请缩小参数参数特征或将配置文件中的" \
                                                                "max_column_length增大".format(table['url'])
                assert len(''.join(
                    tokens)) <= max_query_length, '问句: {0} 输入问句长度大于最大字符长度，请将配置文件中的max_query_length' \
                                                  '增大或者缩短输入问句长度'.format(example.question)
                tokenize_result = self.tokenizer.encode_plus(
                    table_column_feature,
                    tokens,
                    max_length=max_total_length,
                    truncation_strategy="longest_first",
                    pad_to_max_length=True,
                )

                input_ids = tokenize_result["input_ids"]  # 将输入tokens, 通过词表转换成id
                segment_ids = tokenize_result["token_type_ids"]  # 标识column和question位置[0,0,1,1]
                input_mask = tokenize_result["attention_mask"]  # 标识column+question位置编码,也就是整个输入所占的编码
                new_tokens = self.tokenizer.convert_ids_to_tokens(input_ids)  # 获取编码后的序列[cls]column[SEP]question[SEP]

                # 0 1 划分方式
                # [cls] column [sep] question [sep]
                #   0     0      0      1       1
                # 获取column+[sep]的长度
                column_token_length = 0
                for i, sid in enumerate(segment_ids):
                    if sid == 1:
                        column_token_length = i
                        break

                # 计算编码后的subword_to_word，需要加上column_token_length偏移
                subword_to_word = [0] * column_token_length + subword_to_word_base
                # 计算编码后的word_to_subword，需要加上column_token_length偏移
                word_to_subword = [(pos[0] + column_token_length, pos[1] + column_token_length) for pos in
                                   word_to_subword_base]

                assert len(input_ids) == max_total_length
                assert len(input_mask) == max_total_length
                assert len(segment_ids) == max_total_length

                input_feature.tokens.append(new_tokens)
                input_feature.word_to_subword.append(word_to_subword)
                input_feature.subword_to_word.append(subword_to_word)
                input_feature.input_ids.append(input_ids)
                input_feature.input_mask.append(input_mask)
                input_feature.segment_ids.append(segment_ids)

        return input_feature

    # 填充特征
    def fill_label_column_feature_split(self, example: SQLExample, input_feature: ColumnInputFeature,
                                        schema: schema_parsed, config,
                                        positive=True, negative=True):
        """
            example:
            {
                qid,
                seq,
                agg,
                sel,
                conds,
                value_start_end
                relationship
            }
            input_feature:
            {
                tokens
                word_to_subword
                subword_to_word
                input_ids
                input_mask
                segment_ids
            }
        """
        max_query_length = int(config["max_query_length"])  # 最长query长度
        max_column_length = int(config["max_column_length"])  # 最长列长度
        max_total_length = max_column_length + max_query_length  # 最长序列长度

        # set relative column
        table_id = example.table_id
        # 只要正例
        if positive and not negative:
            global_sel_id = example.sel
            global_where_id = [cond[0] for cond in example.conds]
        # 只要负例
        elif not positive and negative:
            global_sel_id = []
            global_where_id = []
        # 正例和负例都要
        elif positive and negative:
            global_sel_id = [schema.local_to_global[str(table_id) + '/' + str(i)] for i in
                             example.sel]
            global_where_id = [schema.local_to_global[str(table_id) + '/' + str(cond[0])] for cond in
                               example.conds]
        else:
            assert False, 'Positive 和 Negative 不能全为负'

        # set column number
        col_num = len(input_feature.input_ids)  # 列数量

        # set seq
        input_feature.seq = [example.seq] * col_num
        # for i in global_sql_id:
        #     input_feature.seq[i] = example.seq

        # set sel_num
        input_feature.sel_num = [0] * col_num
        for i in global_sel_id:
            input_feature.sel_num[i] = len(example.sel)

        # input_feature.sel_num = [len(example.sel)] * col_num

        # set sel
        input_feature.sel = [0] * col_num
        for i in global_sel_id:
            input_feature.sel[i] = 1

        # set agg
        input_feature.agg = [0] * col_num
        for agg_index, global_col_index in enumerate(global_sel_id):
            input_feature.agg[global_col_index] = example.agg[agg_index]
        # for agg_index in range(0, len(example.sel)):
        #     local_id = example.sel[agg_index]
        #     global_id = schema.local_to_global[str(table_id) + '/' + str(local_id)]
        #     input_feature.agg[global_id] = example.agg[agg_index]

        # set where_num
        input_feature.where_num = [0] * col_num
        for i in global_where_id:
            input_feature.where_num[i] = len(example.conds)
        # input_feature.where_num = [len(example.conds)] * col_num

        # set relationship
        input_feature.relationship = [example.relationship] * col_num

        # init where op value_num
        input_feature.where = [0] * col_num
        input_feature.op = [0] * col_num
        input_feature.value_num = [0] * col_num

        # inti value_start_end
        input_feature.value_start = [[0] * int(config['max_value_num']) for i in range(0, col_num)]
        input_feature.value_end = [[0] * int(config['max_value_num']) for i in range(0, col_num)]

        # 只有正例需要
        if positive:
            # set where op value_num value_start_end
            for index, value in enumerate(example.conds):
                global_id = global_where_id[index]  # global id
                col_idx, op, value_num, value_span = value
                input_feature.where[global_id] = 1
                input_feature.op[global_id] = op
                input_feature.value_num[global_id] = value_num

                # 如果column没有对应的值，就跳过
                if len(value_span) == 0:
                    continue
                where_col_value_list = value_span
                # 遍历值
                for value_index, where_col_value in enumerate(where_col_value_list):
                    se = example.value_start_end[where_col_value]  # 获取value的value_start_end
                    try:
                        s = input_feature.word_to_subword[global_id][se[0]][0]  # 获取在编码过后的start_token
                        input_feature.value_start[global_id][value_index] = s
                        # -1 有可能导致大问题，后面看怎么用, 与span那里的l-1对应上了
                        # 这里-1 是预测的开始位置token，与word_to_subword和subword_to_word相对应
                        e = input_feature.word_to_subword[global_id][se[1]][1] - 1  # 获取在编码过后的end_token
                        input_feature.value_end[global_id][value_index] = e

                        # 断言，start_token 和 end_token 的位置均少于编码长度，并且对应input_mask位置的一个token
                        assert s < max_total_length and input_feature.input_mask[global_id][s] == 1
                        assert e < max_total_length and input_feature.input_mask[global_id][e] == 1
                    except:
                        print(example.qid, where_col_value)
                        print("value span is out of range")
                        return False
        return True

    # 加载数据
    def load_data(self, data_paths, config, schema, include_label=False, positive=True, negative=True, note=False):
        '''
            data_paths: 数据路径
            config: 模型参数设置
            include_label: 是否包含标记label
        '''
        # 将input,封装到model_input 中
        # 每一条数据是<>col<>question<>的特征
        model_inputs = {k: [] for k in ["input_ids", "input_mask", "segment_ids"]}
        # 拓展
        if include_label:
            for k in ["seq", "sel_num", "sel", "agg", "where_num",
                      "where", "op", "value_num", "value_start", "value_end", "relationship"]:
                model_inputs[k] = []

        # 记录输入每一个question所对应的特征序列
        pos = []
        # 多条<>column<>question<>特征
        input_features = []

        # 支持多个数据集
        for data_path in data_paths.split("|"):
            cnt = 0
            for line in open(data_path, encoding="utf8"):
                # 加载一个样本, 一行里面包含一个问句，一个问句有可能对应两个SQL，按SQL进行拆分，拆分成两个样本
                # 将所有正负样本标记在一起
                for example in SQLExample.load_from_json(line):
                    # 丢弃未经过验证并且包含sql_label的数据
                    if not example.valid and include_label == True:
                        continue

                    # 获取输入特征,返回一个对象，里面封装了多个<col,question>,<col,question>
                    input_feature = self.get_input_column_feature(example, schema, config, positive, negative, note)

                    # 如果包含sql_label, 则需要填充特征
                    if include_label:
                        success = self.fill_label_column_feature_split(example, input_feature, schema, config, positive,
                                                                       negative)
                        if not success:
                            continue

                    input_features.append(input_feature)  # 总特征

                    # 当前模型有多少个特征
                    cur_start = len(model_inputs["input_ids"])
                    # 当前样本输入有多少个输入特征
                    cur_sample_num = len(input_feature.input_ids)
                    # 标记当前输入模型对应特征的开始和结束位置
                    pos.append((cur_start, cur_start + cur_sample_num))

                    model_inputs["input_ids"].extend(input_feature.input_ids)
                    model_inputs["input_mask"].extend(input_feature.input_mask)
                    model_inputs["segment_ids"].extend(input_feature.segment_ids)
                    if include_label:
                        model_inputs['seq'].extend(input_feature.seq)
                        model_inputs['sel_num'].extend(input_feature.sel_num)
                        model_inputs["sel"].extend(input_feature.sel)
                        model_inputs["agg"].extend(input_feature.agg)
                        model_inputs["where_num"].extend(input_feature.where_num)
                        model_inputs["where"].extend(input_feature.where)
                        model_inputs["op"].extend(input_feature.op)
                        model_inputs["value_num"].extend(input_feature.value_num)
                        model_inputs["value_start"].extend(input_feature.value_start)
                        model_inputs["value_end"].extend(input_feature.value_end)
                        model_inputs["relationship"].extend(input_feature.relationship)
                    else:
                        # 如果 include_label is False，则说明这是预测时用的数据，那么只需要一条特征就行了
                        # 因为可以对一条特征进行两次预测，分别是SQL1 和 SQL2
                        break
                    cnt += 1
                    if cnt % 5000 == 0:
                        print('featuring:', cnt)

        '''
            input_features : 数据的特征，可以认为是label
            model_inputs : 数据输入
            pos :  一条数据与多条<>column<>question<>的对应关系
        '''
        return input_features, model_inputs, pos

    # 加载数据
    def load_predict_data(self, example: SQLExample, config, schema, note=True):
        # 将example,封装到model_input 中
        # 每一条数据是<>col<>question<>的特征
        model_inputs = {k: [] for k in ["input_ids", "input_mask", "segment_ids"]}
        # 记录输入每一个question所对应的特征序列
        pos = []
        # 多条<>column<>question<>特征
        input_features = []
        for table_id in example.first_schema_id + example.sec_schema_id:
            example.table_id = table_id
            # 获取输入特征,返回一个对象，里面封装了多个<col,question>,<col,question>
            input_feature = self.get_predict_input_column_feature(example, schema, config, note)

            input_features.append(input_feature)  # 总特征

            # 当前模型有多少个特征
            cur_start = len(model_inputs["input_ids"])
            # 当前样本输入有多少个输入特征
            cur_sample_num = len(input_feature.input_ids)
            # 标记当前输入模型对应特征的开始和结束位置
            pos.append((cur_start, cur_start + cur_sample_num))

            model_inputs["input_ids"].extend(input_feature.input_ids)
            model_inputs["input_mask"].extend(input_feature.input_mask)
            model_inputs["segment_ids"].extend(input_feature.segment_ids)

            '''
                input_features : 数据的特征，可以认为是label
                model_inputs : 数据输入
                pos :  一条数据与多条<>column<>question<>的对应关系
            '''
        return input_features, model_inputs, pos


class ColumnDataset(torch_data.Dataset):
    def __init__(self, config, featurizer, schema):
        self.config = config
        self.featurizer = featurizer
        self.schema = schema
        self.input_features = None
        self.model_inputs = None
        self.pos = None

    def loads(self, data_paths, include_label=False, positive=True, negative=True, note=False):
        self.input_features, self.model_inputs, self.pos = self.featurizer.load_data(data_paths, self.config,
                                                                                     self.schema, include_label,
                                                                                     positive, negative, note)
        self.trans_data_type()
        print("file {0} loaded".format(data_paths))

    def load_for_predict(self, example: SQLExample, note=True):
        self.input_features, self.model_inputs, self.pos = self.featurizer.load_predict_data(example, self.config,
                                                                                             self.schema, note)
        self.trans_data_type()

    # 转换数据类型
    def trans_data_type(self):
        for k in self.model_inputs:
            self.model_inputs[k] = np.array(self.model_inputs[k], dtype=np.int64)

    # 这两个函数是dataloader的关键
    def __len__(self):
        return self.model_inputs["input_ids"].shape[0]

    def __getitem__(self, idx):
        return {k: v[idx] for k, v in self.model_inputs.items()}


if __name__ == "__main__":
    ''
